# Licensed under a 3-clause BSD style license - see PYFITS.rst

import errno
import gzip
import io
import mmap
import os
import pathlib
import shutil
import sys
import urllib.request
import zipfile
from unittest.mock import patch

import numpy as np
import pytest

from astropy.io import fits
from astropy.io.fits.convenience import _getext
from astropy.io.fits.diff import FITSDiff
from astropy.io.fits.file import GZIP_MAGIC, _File
from astropy.io.tests import safeio
from astropy.utils import data
from astropy.utils.compat.optional_deps import (
    HAS_BZ2,  # NOTE: Python can be built without bz2
    HAS_LZMA,  # NOTE: Python can be built without lzma
    HAS_UNCOMPRESSPY,
)
from astropy.utils.data import conf
from astropy.utils.exceptions import AstropyUserWarning
from astropy.utils.misc import _NOT_OVERWRITING_MSG_MATCH

from .conftest import FitsTestCase

if HAS_BZ2:
    import bz2

if HAS_LZMA:
    import lzma

if HAS_UNCOMPRESSPY:
    import uncompresspy


class TestCore(FitsTestCase):
    def test_missing_file(self):
        with pytest.raises(OSError):
            fits.open(self.temp("does-not-exist.fits"))

    def test_naxisj_check(self):
        with fits.open(self.data("o4sp040b0_raw.fits")) as hdulist:
            hdulist[1].header["NAXIS3"] = 500

            assert "NAXIS3" in hdulist[1].header
            hdulist.verify("silentfix")
            assert "NAXIS3" not in hdulist[1].header

    def test_byteswap(self):
        p = fits.PrimaryHDU()
        lst = fits.HDUList()

        n = np.array([1, 60000, 0], dtype="u2").astype("i2")

        c = fits.Column(name="foo", format="i2", bscale=1, bzero=32768, array=n)
        t = fits.BinTableHDU.from_columns([c])

        lst.append(p)
        lst.append(t)

        lst.writeto(self.temp("test.fits"), overwrite=True)

        with fits.open(self.temp("test.fits")) as p:
            assert p[1].data[1]["foo"] == 60000.0

    def test_fits_file_path_object(self):
        """
        Testing when fits file is passed as pathlib.Path object #4412.
        """
        fpath = pathlib.Path(self.data("tdim.fits"))
        with fits.open(fpath) as hdulist:
            assert hdulist[0].filebytes() == 2880
            assert hdulist[1].filebytes() == 5760

            with fits.open(self.data("tdim.fits")) as hdulist2:
                assert FITSDiff(hdulist2, hdulist).identical is True

    def test_fits_pathlike_object(self):
        """
        Testing when fits file is passed as os.PathLike object #11579.
        """

        class TPath(os.PathLike):
            def __init__(self, path):
                self.path = path

            def __fspath__(self):
                return str(self.path)

        fpath = TPath(self.data("tdim.fits"))
        with fits.open(fpath) as hdulist:
            assert hdulist[0].filebytes() == 2880
            assert hdulist[1].filebytes() == 5760

            with fits.open(self.data("tdim.fits")) as hdulist2:
                assert FITSDiff(hdulist2, hdulist).identical is True

    def test_fits_file_bytes_object(self):
        """
        Testing when fits file is passed as bytes.
        """
        with fits.open(self.data("tdim.fits").encode()) as hdulist:
            assert hdulist[0].filebytes() == 2880
            assert hdulist[1].filebytes() == 5760

            with fits.open(self.data("tdim.fits")) as hdulist2:
                assert FITSDiff(hdulist2, hdulist).identical is True

    def test_add_del_columns(self):
        p = fits.ColDefs([])
        p.add_col(fits.Column(name="FOO", format="3J"))
        p.add_col(fits.Column(name="BAR", format="1I"))
        assert p.names == ["FOO", "BAR"]
        p.del_col("FOO")
        assert p.names == ["BAR"]

    def test_add_del_columns2(self):
        hdulist = fits.open(self.data("tb.fits"))
        table = hdulist[1]
        assert table.data.dtype.names == ("c1", "c2", "c3", "c4")
        assert table.columns.names == ["c1", "c2", "c3", "c4"]
        table.columns.del_col("c1")
        assert table.data.dtype.names == ("c2", "c3", "c4")
        assert table.columns.names == ["c2", "c3", "c4"]

        table.columns.del_col("c3")
        assert table.data.dtype.names == ("c2", "c4")
        assert table.columns.names == ["c2", "c4"]

        table.columns.add_col(fits.Column("foo", "3J"))
        assert table.data.dtype.names == ("c2", "c4", "foo")
        assert table.columns.names == ["c2", "c4", "foo"]

        hdulist.writeto(self.temp("test.fits"), overwrite=True)
        hdulist.close()
        # NOTE: If you see a warning, might be related to
        # https://github.com/spacetelescope/PyFITS/issues/44
        with fits.open(self.temp("test.fits")) as hdulist:
            table = hdulist[1]
            assert table.data.dtype.names == ("c2", "c4", "foo")
            assert table.columns.names == ["c2", "c4", "foo"]

    def test_update_header_card(self):
        """A very basic test for the Header.update method--I'd like to add a
        few more cases to this at some point.
        """

        header = fits.Header()
        comment = "number of bits per data pixel"
        header["BITPIX"] = (16, comment)
        assert "BITPIX" in header
        assert header["BITPIX"] == 16
        assert header.comments["BITPIX"] == comment

        header.update(BITPIX=32)
        assert header["BITPIX"] == 32
        assert header.comments["BITPIX"] == ""

    def test_set_card_value(self):
        """Similar to test_update_header_card(), but tests the
        `header['FOO'] = 'bar'` method of updating card values.
        """

        header = fits.Header()
        comment = "number of bits per data pixel"
        card = fits.Card.fromstring(f"BITPIX  = 32 / {comment}")
        header.append(card)

        header["BITPIX"] = 32

        assert "BITPIX" in header
        assert header["BITPIX"] == 32
        assert header.cards[0].keyword == "BITPIX"
        assert header.cards[0].value == 32
        assert header.cards[0].comment == comment

    def test_uint(self):
        filename = self.data("o4sp040b0_raw.fits")
        with fits.open(filename, uint=False) as hdulist_f:
            with fits.open(filename, uint=True) as hdulist_i:
                assert hdulist_f[1].data.dtype == np.float32
                assert hdulist_i[1].data.dtype == np.uint16
                assert np.all(hdulist_f[1].data == hdulist_i[1].data)

    def test_fix_missing_card_append(self):
        hdu = fits.ImageHDU()
        errs = hdu.req_cards("TESTKW", None, None, "foo", "silentfix", [])
        assert len(errs) == 1
        assert "TESTKW" in hdu.header
        assert hdu.header["TESTKW"] == "foo"
        assert hdu.header.cards[-1].keyword == "TESTKW"

    def test_fix_invalid_keyword_value(self):
        hdu = fits.ImageHDU()
        hdu.header["TESTKW"] = "foo"
        errs = hdu.req_cards("TESTKW", None, lambda v: v == "foo", "foo", "ignore", [])
        assert len(errs) == 0

        # Now try a test that will fail, and ensure that an error will be
        # raised in 'exception' mode
        errs = hdu.req_cards(
            "TESTKW", None, lambda v: v == "bar", "bar", "exception", []
        )
        assert len(errs) == 1
        assert errs[0][1] == "'TESTKW' card has invalid value 'foo'."

        # See if fixing will work
        hdu.req_cards("TESTKW", None, lambda v: v == "bar", "bar", "silentfix", [])
        assert hdu.header["TESTKW"] == "bar"

    def test_unfixable_missing_card(self):
        class TestHDU(fits.hdu.base.NonstandardExtHDU):
            def _verify(self, option="warn"):
                errs = super()._verify(option)
                hdu.req_cards("TESTKW", None, None, None, "fix", errs)
                return errs

            @classmethod
            def match_header(cls, header):
                # Since creating this HDU class adds it to the registry we
                # don't want the file reader to possibly think any actual
                # HDU from a file should be handled by this class
                return False

        hdu = TestHDU(header=fits.Header())
        with pytest.raises(fits.VerifyError):
            hdu.verify("fix")

    def test_exception_on_verification_error(self):
        hdu = fits.ImageHDU()
        del hdu.header["XTENSION"]
        with pytest.raises(fits.VerifyError):
            hdu.verify("exception")

    def test_ignore_verification_error(self):
        hdu = fits.ImageHDU()
        del hdu.header["NAXIS"]
        # The default here would be to issue a warning; ensure that no warnings
        # or exceptions are raised
        hdu.verify("ignore")

        # Make sure the error wasn't fixed either, silently or otherwise
        assert "NAXIS" not in hdu.header

    def test_unrecognized_verify_option(self):
        hdu = fits.ImageHDU()
        with pytest.raises(ValueError):
            hdu.verify("foobarbaz")

    def test_errlist_basic(self):
        # Just some tests to make sure that _ErrList is setup correctly.
        # No arguments
        error_list = fits.verify._ErrList()
        assert error_list == []
        # Some contents - this is not actually working, it just makes sure they
        # are kept.
        error_list = fits.verify._ErrList([1, 2, 3])
        assert error_list == [1, 2, 3]

    def test_combined_verify_options(self):
        """
        Test verify options like fix+ignore.
        """

        def make_invalid_hdu():
            hdu = fits.ImageHDU()
            # Add one keyword to the header that contains a fixable defect, and one
            # with an unfixable defect
            c1 = fits.Card.fromstring("test    = '    test'")
            c2 = fits.Card.fromstring("P.I.    = '  Hubble'")
            hdu.header.append(c1)
            hdu.header.append(c2)
            return hdu

        # silentfix+ignore should be completely silent
        hdu = make_invalid_hdu()
        hdu.verify("silentfix+ignore")

        # silentfix+warn should be quiet about the fixed HDU and only warn
        # about the unfixable one
        hdu = make_invalid_hdu()
        with pytest.warns(fits.verify.VerifyWarning) as w:
            hdu.verify("silentfix+warn")
        assert len(w) == 4
        assert "Illegal keyword name 'P.I.'" in str(w.list[2].message)

        # silentfix+exception should only mention the unfixable error in the
        # exception
        hdu = make_invalid_hdu()
        with pytest.raises(fits.VerifyError, match=r"Illegal keyword name") as excinfo:
            hdu.verify("silentfix+exception")
        assert "not upper case" not in str(excinfo.value)

        # fix+ignore is not too useful, but it should warn about the fixed
        # problems while saying nothing about the unfixable problems
        hdu = make_invalid_hdu()
        with pytest.warns(fits.verify.VerifyWarning) as w:
            hdu.verify("fix+ignore")
        assert len(w) == 4
        assert "not upper case" not in str(w.list[0].message)

        # fix+warn
        hdu = make_invalid_hdu()
        with pytest.warns(AstropyUserWarning) as w:
            hdu.verify("fix+warn")
        assert len(w) == 6
        assert "not upper case" in str(w[2].message)
        assert "Illegal keyword name" in str(w[4].message)

        # fix+exception
        hdu = make_invalid_hdu()
        with pytest.raises(fits.VerifyError, match=r"Illegal keyword name") as excinfo:
            hdu.verify("fix+exception")
        assert "not upper case" in str(excinfo.value)

    def test_getext(self):
        """
        Test the various different ways of specifying an extension header in
        the convenience functions.
        """
        filename = self.data("test0.fits")

        hl, ext = _getext(filename, "readonly", 1)
        assert ext == 1
        hl.close()

        pytest.raises(ValueError, _getext, filename, "readonly", 1, 2)
        pytest.raises(ValueError, _getext, filename, "readonly", (1, 2))
        pytest.raises(ValueError, _getext, filename, "readonly", "sci", "sci")
        pytest.raises(TypeError, _getext, filename, "readonly", 1, 2, 3)

        hl, ext = _getext(filename, "readonly", ext=1)
        assert ext == 1
        hl.close()

        hl, ext = _getext(filename, "readonly", ext=("sci", 2))
        assert ext == ("sci", 2)
        hl.close()

        pytest.raises(
            TypeError, _getext, filename, "readonly", 1, ext=("sci", 2), extver=3
        )
        pytest.raises(
            TypeError, _getext, filename, "readonly", ext=("sci", 2), extver=3
        )

        hl, ext = _getext(filename, "readonly", "sci")
        assert ext == ("sci", 1)
        hl.close()

        hl, ext = _getext(filename, "readonly", "sci", 1)
        assert ext == ("sci", 1)
        hl.close()

        hl, ext = _getext(filename, "readonly", ("sci", 1))
        assert ext == ("sci", 1)
        hl.close()

        hl, ext = _getext(
            filename, "readonly", "sci", extver=1, do_not_scale_image_data=True
        )
        assert ext == ("sci", 1)
        hl.close()

        pytest.raises(TypeError, _getext, filename, "readonly", "sci", ext=1)
        pytest.raises(TypeError, _getext, filename, "readonly", "sci", 1, extver=2)

        hl, ext = _getext(filename, "readonly", extname="sci")
        assert ext == ("sci", 1)
        hl.close()

        hl, ext = _getext(filename, "readonly", extname="sci", extver=1)
        assert ext == ("sci", 1)
        hl.close()

        pytest.raises(TypeError, _getext, filename, "readonly", extver=1)

    def test_extension_name_case_sensitive(self):
        """
        Tests that setting fits.conf.extension_name_case_sensitive at
        runtime works.
        """

        hdu = fits.ImageHDU()
        hdu.name = "sCi"
        assert hdu.name == "SCI"
        assert hdu.header["EXTNAME"] == "SCI"

        with fits.conf.set_temp("extension_name_case_sensitive", True):
            hdu = fits.ImageHDU()
            hdu.name = "sCi"
            assert hdu.name == "sCi"
            assert hdu.header["EXTNAME"] == "sCi"

        hdu.name = "sCi"
        assert hdu.name == "SCI"
        assert hdu.header["EXTNAME"] == "SCI"

    def test_hdu_fromstring(self):
        """
        Tests creating a fully-formed HDU object from a string containing the
        bytes of the HDU.
        """
        infile = self.data("test0.fits")
        outfile = self.temp("test.fits")

        with open(infile, "rb") as fin:
            dat = fin.read()

        offset = 0
        with fits.open(infile) as hdul:
            hdulen = hdul[0]._data_offset + hdul[0]._data_size
            hdu = fits.PrimaryHDU.fromstring(dat[:hdulen])
            assert isinstance(hdu, fits.PrimaryHDU)
            assert hdul[0].header == hdu.header
            assert hdu.data is None

        hdu.header["TEST"] = "TEST"
        hdu.writeto(outfile)
        with fits.open(outfile) as hdul:
            assert isinstance(hdu, fits.PrimaryHDU)
            assert hdul[0].header[:-1] == hdu.header[:-1]
            assert hdul[0].header["TEST"] == "TEST"
            assert hdu.data is None

        with fits.open(infile) as hdul:
            for ext_hdu in hdul[1:]:
                offset += hdulen
                hdulen = len(str(ext_hdu.header)) + ext_hdu._data_size
                hdu = fits.ImageHDU.fromstring(dat[offset : offset + hdulen])
                assert isinstance(hdu, fits.ImageHDU)
                assert ext_hdu.header == hdu.header
                assert (ext_hdu.data == hdu.data).all()

    def test_nonstandard_hdu(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/157

        Tests that "Nonstandard" HDUs with SIMPLE = F are read and written
        without prepending a superfluous and unwanted standard primary HDU.
        """

        data = np.arange(100, dtype=np.uint8)
        hdu = fits.PrimaryHDU(data=data)
        hdu.header["SIMPLE"] = False
        hdu.writeto(self.temp("test.fits"))

        info = [(0, "", 1, "NonstandardHDU", 5, (), "", "")]
        with fits.open(self.temp("test.fits")) as hdul:
            assert hdul.info(output=False) == info
            # NonstandardHDUs just treat the data as an unspecified array of
            # bytes.  The first 100 bytes should match the original data we
            # passed in...the rest should be zeros padding out the rest of the
            # FITS block
            assert (hdul[0].data[:100] == data).all()
            assert (hdul[0].data[100:] == 0).all()

    def test_extname(self):
        """Test getting/setting the EXTNAME of an HDU."""

        h1 = fits.PrimaryHDU()
        assert h1.name == "PRIMARY"
        # Normally a PRIMARY HDU should not have an EXTNAME, though it should
        # have a default .name attribute
        assert "EXTNAME" not in h1.header

        # The current version of the FITS standard does allow PRIMARY HDUs to
        # have an EXTNAME, however.
        h1.name = "NOTREAL"
        assert h1.name == "NOTREAL"
        assert h1.header.get("EXTNAME") == "NOTREAL"

        # Updating the EXTNAME in the header should update the .name
        h1.header["EXTNAME"] = "TOOREAL"
        assert h1.name == "TOOREAL"

        # If we delete an EXTNAME keyword from a PRIMARY HDU it should go back
        # to the default
        del h1.header["EXTNAME"]
        assert h1.name == "PRIMARY"

        # For extension HDUs the situation is a bit simpler:
        h2 = fits.ImageHDU()
        assert h2.name == ""
        assert "EXTNAME" not in h2.header
        h2.name = "HELLO"
        assert h2.name == "HELLO"
        assert h2.header.get("EXTNAME") == "HELLO"
        h2.header["EXTNAME"] = "GOODBYE"
        assert h2.name == "GOODBYE"

    def test_extver_extlevel(self):
        """Test getting/setting the EXTVER and EXTLEVEL of and HDU."""

        # EXTVER and EXTNAME work exactly the same; their semantics are, for
        # now, to be inferred by the user.  Although they should never be less
        # than 1, the standard does not explicitly forbid any value so long as
        # it's an integer
        h1 = fits.PrimaryHDU()
        assert h1.ver == 1
        assert h1.level == 1
        assert "EXTVER" not in h1.header
        assert "EXTLEVEL" not in h1.header

        h1.ver = 2
        assert h1.header.get("EXTVER") == 2
        h1.header["EXTVER"] = 3
        assert h1.ver == 3
        del h1.header["EXTVER"]
        assert h1.ver == 1

        h1.level = 2
        assert h1.header.get("EXTLEVEL") == 2
        h1.header["EXTLEVEL"] = 3
        assert h1.level == 3
        del h1.header["EXTLEVEL"]
        assert h1.level == 1

        pytest.raises(TypeError, setattr, h1, "ver", "FOO")
        pytest.raises(TypeError, setattr, h1, "level", "BAR")

    def test_consecutive_writeto(self):
        """
        Regression test for an issue where calling writeto twice on the same
        HDUList could write a corrupted file.

        https://github.com/spacetelescope/PyFITS/issues/40 is actually a
        particular instance of this problem, though isn't unique to sys.stdout.
        """

        with fits.open(self.data("test0.fits")) as hdul1:
            # Add a bunch of header keywords so that the data will be forced to
            # new offsets within the file:
            for idx in range(40):
                hdul1[1].header[f"TEST{idx}"] = "test"

            hdul1.writeto(self.temp("test1.fits"))
            hdul1.writeto(self.temp("test2.fits"))

            # Open a second handle to the original file and compare it to hdul1
            # (We only compare part of the one header that was modified)
            # Compare also with the second writeto output
            with fits.open(self.data("test0.fits")) as hdul2:
                with fits.open(self.temp("test2.fits")) as hdul3:
                    for hdul in (hdul1, hdul3):
                        for idx, hdus in enumerate(zip(hdul2, hdul)):
                            hdu2, hdu = hdus
                            if idx != 1:
                                assert hdu.header == hdu2.header
                            else:
                                assert hdu2.header == hdu.header[: len(hdu2.header)]
                            assert np.all(hdu.data == hdu2.data)


class TestConvenienceFunctions(FitsTestCase):
    def test_writeto(self, home_is_temp):
        """
        Simple test for writing a trivial header and some data to a file
        with the `writeto()` convenience function.
        """
        filename = self.temp("array.fits")
        data = np.zeros((100, 100))
        header = fits.Header()
        fits.writeto(filename, data, header=header, overwrite=True)
        with fits.open(filename) as hdul:
            assert len(hdul) == 1
            assert (data == hdul[0].data).all()

    def test_writeto_2(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/107

        Test of `writeto()` with a trivial header containing a single keyword.
        """
        filename = self.temp("array.fits")
        data = np.zeros((100, 100))
        header = fits.Header()
        header.set("CRPIX1", 1.0)
        fits.writeto(
            filename, data, header=header, overwrite=True, output_verify="silentfix"
        )
        with fits.open(filename) as hdul:
            assert len(hdul) == 1
            assert (data == hdul[0].data).all()
            assert "CRPIX1" in hdul[0].header
            assert hdul[0].header["CRPIX1"] == 1.0

    def test_writeto_overwrite(self, home_is_temp):
        """
        Ensure the `overwrite` keyword works as it should
        """
        filename = self.temp("array.fits")
        data = np.zeros((100, 100))
        header = fits.Header()
        fits.writeto(filename, data, header=header)

        with pytest.raises(OSError, match=_NOT_OVERWRITING_MSG_MATCH):
            fits.writeto(filename, data, header=header, overwrite=False)

        fits.writeto(filename, data, header=header, overwrite=True)

        with fits.open(filename) as hdul:
            assert len(hdul) == 1
            assert (data == hdul[0].data).all()

    def test_writeto_stdout(self):
        # see https://github.com/astropy/astropy/issues/3427
        fits.writeto(sys.stdout, data=np.array([1, 2]))


class TestFileFunctions(FitsTestCase):
    """
    Tests various basic I/O operations, specifically in the
    astropy.io.fits.file._File class.
    """

    def test_open_nonexistent(self):
        """Test that trying to open a non-existent file results in an
        OSError (and not some other arbitrary exception).
        """

        with pytest.raises(OSError, match=r"No such file or directory"):
            fits.open(self.temp("foobar.fits"))

        # But opening in ostream or append mode should be okay, since they
        # allow writing new files
        for mode in ("ostream", "append"):
            with fits.open(self.temp("foobar.fits"), mode=mode) as _:
                pass

            assert os.path.exists(self.temp("foobar.fits"))
            os.remove(self.temp("foobar.fits"))

    def test_open_file_handle(self):
        # Make sure we can open a FITS file from an open file handle
        with open(self.data("test0.fits"), "rb") as handle:
            with fits.open(handle) as _:
                pass

        with open(self.temp("temp.fits"), "wb") as handle:
            with fits.open(handle, mode="ostream") as _:
                pass

        # Opening without explicitly specifying binary mode should fail
        with pytest.raises(ValueError):
            with open(self.data("test0.fits")) as handle:
                with fits.open(handle) as _:
                    pass

        # All of these read modes should fail
        for mode in ["r", "rt"]:
            with pytest.raises(ValueError):
                with open(self.data("test0.fits"), mode=mode) as handle:
                    with fits.open(handle) as _:
                        pass

        # These update or write modes should fail as well
        for mode in ["w", "wt", "w+", "wt+", "r+", "rt+", "a", "at", "a+", "at+"]:
            with pytest.raises(ValueError):
                with open(self.temp("temp.fits"), mode=mode) as handle:
                    with fits.open(handle) as _:
                        pass

    def test_fits_file_handle_mode_combo(self):
        # This should work fine since no mode is given
        with open(self.data("test0.fits"), "rb") as handle:
            with fits.open(handle) as _:
                pass

        # This should work fine since the modes are compatible
        with open(self.data("test0.fits"), "rb") as handle:
            with fits.open(handle, mode="readonly") as _:
                pass

        # This should not work since the modes conflict
        with pytest.raises(ValueError):
            with open(self.data("test0.fits"), "rb") as handle:
                with fits.open(handle, mode="ostream") as _:
                    pass

    def test_open_from_url(self):
        file_url = "file:///" + self.data("test0.fits").lstrip("/")
        with urllib.request.urlopen(file_url) as urlobj:
            with fits.open(urlobj) as _:
                pass

        # It will not be possible to write to a file that is from a URL object
        for mode in ("ostream", "append", "update"):
            with pytest.raises(ValueError):
                with urllib.request.urlopen(file_url) as urlobj:
                    with fits.open(urlobj, mode=mode) as _:
                        pass

    @pytest.mark.remote_data(source="astropy")
    def test_open_from_remote_url(self):
        for dataurl in (conf.dataurl, conf.dataurl_mirror):
            remote_url = f"{dataurl}/allsky/allsky_rosat.fits"
            try:
                with urllib.request.urlopen(remote_url) as urlobj:
                    with fits.open(urlobj) as fits_handle:
                        assert len(fits_handle) == 1

                for mode in ("ostream", "append", "update"):
                    with pytest.raises(ValueError):
                        with urllib.request.urlopen(remote_url) as urlobj:
                            with fits.open(urlobj, mode=mode) as fits_handle:
                                assert len(fits_handle) == 1
            except (urllib.error.HTTPError, urllib.error.URLError):
                continue
            else:
                break
        else:
            raise Exception("Could not download file")

    def test_open_gzipped(self):
        gzip_file = self._make_gzip_file()

        with fits.open(gzip_file) as fits_handle:
            assert fits_handle._file.compression == "gzip"
            assert len(fits_handle) == 5

        with fits.open(gzip_file, decompress_in_memory=True) as fits_handle:
            assert fits_handle._file.compression == "gzip"
            assert len(fits_handle) == 5

        with fits.open(gzip.GzipFile(gzip_file)) as fits_handle:
            assert fits_handle._file.compression == "gzip"
            assert len(fits_handle) == 5

    def test_open_gzipped_from_handle(self):
        with open(self._make_gzip_file(), "rb") as handle:
            with fits.open(handle) as fits_handle:
                assert fits_handle._file.compression == "gzip"

    def test_detect_gzipped(self):
        """Test detection of a gzip file when the extension is not .gz."""
        with fits.open(self._make_gzip_file("test0.fz")) as fits_handle:
            assert fits_handle._file.compression == "gzip"
            assert len(fits_handle) == 5

    def test_writeto_append_mode_gzip(self):
        """Regression test for
        https://github.com/spacetelescope/PyFITS/issues/33

        Check that a new GzipFile opened in append mode can be used to write
        out a new FITS file.
        """

        # Note: when opening a GzipFile the 'b+' is superfluous, but this was
        # still how the original test case looked
        # Note: with statement not supported on GzipFile in older Python
        # versions
        fileobj = gzip.GzipFile(self.temp("test.fits.gz"), "ab+")
        h = fits.PrimaryHDU()
        try:
            h.writeto(fileobj)
        finally:
            fileobj.close()

        with fits.open(self.temp("test.fits.gz")) as hdul:
            assert hdul[0].header == h.header

    def test_fits_update_mode_gzip(self):
        """Test updating a GZipped FITS file"""

        with fits.open(self._make_gzip_file("update.gz"), mode="update") as fits_handle:
            hdu = fits.ImageHDU(data=list(range(100)))
            fits_handle.append(hdu)

        with fits.open(self.temp("update.gz")) as new_handle:
            assert len(new_handle) == 6
            assert (new_handle[-1].data == list(range(100))).all()

    def test_fits_append_mode_gzip(self):
        """Make sure that attempting to open an existing GZipped FITS file in
        'append' mode raises an error"""

        with pytest.raises(OSError):
            with fits.open(self._make_gzip_file("append.gz"), mode="append") as _:
                pass

    @pytest.mark.skipif(not HAS_BZ2, reason="Python built without bz2 module")
    def test_open_bzipped(self):
        bzip_file = self._make_bzip2_file()

        with fits.open(bzip_file) as fits_handle:
            assert fits_handle._file.compression == "bzip2"
            assert len(fits_handle) == 5

        with fits.open(bzip_file, decompress_in_memory=True) as fits_handle:
            assert fits_handle._file.compression == "bzip2"
            assert len(fits_handle) == 5

        with fits.open(bz2.BZ2File(bzip_file)) as fits_handle:
            assert fits_handle._file.compression == "bzip2"
            assert len(fits_handle) == 5

        for mode in ("append", "update"):
            with pytest.raises(
                OSError,
                match="update and append modes are not supported with bzip2 files",
            ):
                with fits.open(bzip_file, mode=mode) as fits_handle:
                    pass

    @pytest.mark.skipif(not HAS_BZ2, reason="Python built without bz2 module")
    def test_open_bzipped_from_handle(self):
        with open(self._make_bzip2_file(), "rb") as handle:
            with fits.open(handle) as fits_handle:
                assert fits_handle._file.compression == "bzip2"
                assert len(fits_handle) == 5

    @pytest.mark.skipif(not HAS_BZ2, reason="Python built without bz2 module")
    def test_detect_bzipped(self):
        """Test detection of a bzip2 file when the extension is not .bz2."""
        with fits.open(self._make_bzip2_file("test0.xx")) as fits_handle:
            assert fits_handle._file.compression == "bzip2"
            assert len(fits_handle) == 5

    @pytest.mark.skipif(not HAS_BZ2, reason="Python built without bz2 module")
    def test_writeto_bzip2_fileobj(self):
        """Test writing to a bz2.BZ2File file like object"""
        fileobj = bz2.BZ2File(self.temp("test.fits.bz2"), "w")
        h = fits.PrimaryHDU()
        try:
            h.writeto(fileobj)
        finally:
            fileobj.close()

        with fits.open(self.temp("test.fits.bz2")) as hdul:
            assert hdul[0].header == h.header

    @pytest.mark.skipif(not HAS_BZ2, reason="Python built without bz2 module")
    def test_writeto_bzip2_filename(self):
        """Test writing to a bzip2 file by name"""
        filename = self.temp("testname.fits.bz2")
        h = fits.PrimaryHDU()
        h.writeto(filename)

        with fits.open(self.temp("testname.fits.bz2")) as hdul:
            assert hdul[0].header == h.header

    @pytest.mark.skipif(not HAS_LZMA, reason="Python built without lzma module")
    def test_open_lzma(self):
        lzma_file = self._make_lzma_file()

        with fits.open(lzma_file) as fits_handle:
            assert fits_handle._file.compression == "lzma"
            assert len(fits_handle) == 5

        with fits.open(lzma_file, decompress_in_memory=True) as fits_handle:
            assert fits_handle._file.compression == "lzma"
            assert len(fits_handle) == 5

        with fits.open(lzma.LZMAFile(lzma_file)) as fits_handle:
            assert fits_handle._file.compression == "lzma"
            assert len(fits_handle) == 5

        for mode in ("append", "update"):
            with pytest.raises(
                OSError,
                match="update and append modes are not supported with lzma files",
            ):
                with fits.open(lzma_file, mode=mode) as fits_handle:
                    pass

    @pytest.mark.skipif(not HAS_LZMA, reason="Python built without lzma module")
    def test_open_lzma_from_handle(self):
        with open(self._make_lzma_file(), "rb") as handle:
            with fits.open(handle) as fits_handle:
                assert fits_handle._file.compression == "lzma"
                assert len(fits_handle) == 5

    @pytest.mark.skipif(not HAS_LZMA, reason="Python built without lzma module")
    def test_detect_lzma(self):
        """Test detection of a lzma file when the extension is not .xz."""
        with fits.open(self._make_lzma_file("test0.xx")) as fits_handle:
            assert fits_handle._file.compression == "lzma"
            assert len(fits_handle) == 5

    @pytest.mark.skipif(not HAS_LZMA, reason="Python built without lzma module")
    def test_writeto_lzma_fileobj(self):
        """Test writing to a lzma.LZMAFile file like object"""
        fileobj = lzma.LZMAFile(self.temp("test.fits.xz"), "w")
        h = fits.PrimaryHDU()
        try:
            h.writeto(fileobj)
        finally:
            fileobj.close()

        with fits.open(self.temp("test.fits.xz")) as hdul:
            assert hdul[0].header == h.header

    @pytest.mark.skipif(not HAS_LZMA, reason="Python built without lzma module")
    def test_writeto_lzma_filename(self):
        """Test writing to a lzma file by name"""
        filename = self.temp("testname.fits.xz")
        h = fits.PrimaryHDU()
        h.writeto(filename)

        with fits.open(self.temp("testname.fits.xz")) as hdul:
            assert hdul[0].header == h.header

    @pytest.mark.skipif(
        not HAS_UNCOMPRESSPY, reason="Optional package uncompresspy not installed"
    )
    def test_open_lzw(self):
        lzw_file = self._make_lzw_file()

        arcfile = "ONTT.1991-12-30T08:55:46.000.fits"
        last_datapoint = 53
        with fits.open(lzw_file) as fits_handle:
            assert fits_handle._file.compression == "lzw"
            assert len(fits_handle) == 1
            assert fits_handle[0].header["ARCFILE"] == arcfile
            assert fits_handle[0].data[-1, -1] == last_datapoint

        with fits.open(lzw_file, decompress_in_memory=True) as fits_handle:
            assert fits_handle._file.compression == "lzw"
            assert len(fits_handle) == 1
            assert fits_handle[0].header["ARCFILE"] == arcfile
            assert fits_handle[0].data[-1, -1] == last_datapoint

        with fits.open(uncompresspy.LZWFile(lzw_file)) as fits_handle:
            assert fits_handle._file.compression == "lzw"
            assert len(fits_handle) == 1
            assert fits_handle[0].header["ARCFILE"] == arcfile
            assert fits_handle[0].data[-1, -1] == last_datapoint

        for mode in ("append", "update"):
            with pytest.raises(
                OSError, match=f"{mode} mode not supported with LZW files"
            ):
                with fits.open(lzw_file, mode=mode) as fits_handle:
                    pass

    @pytest.mark.skipif(
        not HAS_UNCOMPRESSPY, reason="Optional package uncompresspy not installed"
    )
    def test_open_lzw_from_handle(self):
        arcfile = "ONTT.1991-12-30T08:55:46.000.fits"
        last_datapoint = 53
        with open(self._make_lzw_file(), "rb") as handle:
            with fits.open(handle) as fits_handle:
                assert fits_handle._file.compression == "lzw"
                assert len(fits_handle) == 1
                assert fits_handle[0].header["ARCFILE"] == arcfile
                assert fits_handle[0].data[-1, -1] == last_datapoint

    @pytest.mark.skipif(
        not HAS_UNCOMPRESSPY, reason="Optional package uncompresspy not installed"
    )
    def test_detect_lzw(self):
        """Test detection of a lzw file when the extension is not .Z."""
        arcfile = "ONTT.1991-12-30T08:55:46.000.fits"
        last_datapoint = 53
        with fits.open(self._make_lzw_file("test0.xx")) as fits_handle:
            assert fits_handle._file.compression == "lzw"
            assert len(fits_handle) == 1
            assert fits_handle[0].header["ARCFILE"] == arcfile
            assert fits_handle[0].data[-1, -1] == last_datapoint

    @pytest.mark.skipif(
        not HAS_UNCOMPRESSPY, reason="Optional package uncompresspy not installed"
    )
    def test_writeto_lzw_filename(self):
        """Test writing to a LZW file by name. This should fail as writing LZW
        is not supported."""
        filename = self.temp("testname.fits.Z")
        h = fits.PrimaryHDU()
        with pytest.raises(OSError, match="mode not supported with LZW files"):
            h.writeto(filename)

    def test_open_zipped(self):
        zip_file = self._make_zip_file()

        with fits.open(zip_file) as fits_handle:
            assert fits_handle._file.compression == "zip"
            assert len(fits_handle) == 5

        with fits.open(zip_file, decompress_in_memory=True) as fits_handle:
            assert fits_handle._file.compression == "zip"
            assert len(fits_handle) == 5

        with fits.open(zipfile.ZipFile(zip_file)) as fits_handle:
            assert fits_handle._file.compression == "zip"
            assert len(fits_handle) == 5

    def test_open_zipped_from_handle(self):
        with open(self._make_zip_file(), "rb") as handle:
            with fits.open(handle) as fits_handle:
                assert fits_handle._file.compression == "zip"
                assert len(fits_handle) == 5

    def test_detect_zipped(self):
        """Test detection of a zip file when the extension is not .zip."""

        zf = self._make_zip_file(filename="test0.fz")
        with fits.open(zf) as fits_handle:
            assert len(fits_handle) == 5

    def test_open_zipped_writeable(self):
        """Opening zipped files in a writeable mode should fail."""

        zf = self._make_zip_file()
        pytest.raises(OSError, fits.open, zf, "update")
        pytest.raises(OSError, fits.open, zf, "append")

        zf = zipfile.ZipFile(zf, "a")
        pytest.raises(OSError, fits.open, zf, "update")
        pytest.raises(OSError, fits.open, zf, "append")

    def test_read_open_astropy_gzip_file(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2774

        This tests reading from a ``GzipFile`` object from Astropy's
        compatibility copy of the ``gzip`` module.
        """
        gf = gzip.GzipFile(self._make_gzip_file())
        try:
            assert len(fits.open(gf)) == 5
        finally:
            gf.close()

    def test_open_multiple_member_zipfile(self):
        """
        Opening zip files containing more than one member files should fail
        as there's no obvious way to specify which file is the FITS file to
        read.
        """

        zfile = zipfile.ZipFile(self.temp("test0.zip"), "w")
        zfile.write(self.data("test0.fits"))
        zfile.writestr("foo", "bar")
        zfile.close()

        with pytest.raises(OSError):
            fits.open(zfile.filename)

    def test_read_open_file(self):
        """Read from an existing file object."""

        with open(self.data("test0.fits"), "rb") as f:
            assert len(fits.open(f)) == 5

    def test_read_closed_file(self):
        """Read from an existing file object that's been closed."""

        f = open(self.data("test0.fits"), "rb")
        f.close()
        with fits.open(f) as f2:
            assert len(f2) == 5

    def test_read_open_gzip_file(self):
        """Read from an open gzip file object."""

        gf = gzip.GzipFile(self._make_gzip_file())
        try:
            assert len(fits.open(gf)) == 5
        finally:
            gf.close()

    def test_open_gzip_file_for_writing(self):
        """Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/195."""

        gf = self._make_gzip_file()
        with fits.open(gf, mode="update") as h:
            h[0].header["EXPFLAG"] = "ABNORMAL"
            h[1].data[0, 0] = 1
        with fits.open(gf) as h:
            # Just to make sure the update worked; if updates work
            # normal writes should work too...
            assert h[0].header["EXPFLAG"] == "ABNORMAL"
            assert h[1].data[0, 0] == 1

    def test_write_read_gzip_file(self, home_is_temp):
        """
        Regression test for https://github.com/astropy/astropy/issues/2794

        Ensure files written through gzip are readable.
        """

        data = np.arange(100)
        hdu = fits.PrimaryHDU(data=data)
        hdu.writeto(self.temp("test.fits.gz"))

        with open(os.path.expanduser(self.temp("test.fits.gz")), "rb") as f:
            assert f.read(3) == GZIP_MAGIC

        with fits.open(self.temp("test.fits.gz")) as hdul:
            assert np.all(hdul[0].data == data)

    @pytest.mark.parametrize("ext", ["gz", "bz2", "zip", "xz", "Z"])
    def test_compressed_ext_but_not_compressed(self, ext):
        testfile = self.temp(f"test0.fits.{ext}")
        shutil.copy(self.data("test0.fits"), testfile)

        with fits.open(testfile) as hdul:
            assert len(hdul) == 5

        fits.append(testfile, np.arange(5))

        with fits.open(testfile) as hdul:
            assert len(hdul) == 6

    def test_read_file_like_object(self):
        """Test reading a FITS file from a file-like object."""

        filelike = io.BytesIO()
        with open(self.data("test0.fits"), "rb") as f:
            filelike.write(f.read())
        filelike.seek(0)
        assert len(fits.open(filelike)) == 5

    def test_updated_file_permissions(self):
        """
        Regression test for https://aeon.stsci.edu/ssb/trac/pyfits/ticket/79

        Tests that when a FITS file is modified in update mode, the file
        permissions are preserved.
        """

        filename = self.temp("test.fits")
        hdul = [fits.PrimaryHDU(), fits.ImageHDU()]
        hdul = fits.HDUList(hdul)
        hdul.writeto(filename)

        old_mode = os.stat(filename).st_mode

        hdul = fits.open(filename, mode="update")
        hdul.insert(1, fits.ImageHDU())
        hdul.flush()
        hdul.close()

        assert old_mode == os.stat(filename).st_mode

    def test_fileobj_mode_guessing(self):
        """Tests whether a file opened without a specified io.fits mode
        ('readonly', etc.) is opened in a mode appropriate for the given file
        object.
        """

        testfile = self.copy_file("test0.fits")

        # Opening in text mode should outright fail
        for mode in ("r", "w", "a"):
            with open(testfile, mode) as f:
                pytest.raises(ValueError, fits.HDUList.fromfile, f)

        # Need to re-copy the file since opening it in 'w' mode blew it away
        testfile = self.copy_file("test0.fits")

        with open(testfile, "rb") as f:
            with fits.HDUList.fromfile(f) as h:
                assert h.fileinfo(0)["filemode"] == "readonly"

        for mode in ("wb", "ab"):
            with open(testfile, mode) as f:
                with fits.HDUList.fromfile(f) as h:
                    # Basically opening empty files for output streaming
                    assert len(h) == 0

        # Need to re-copy the file since opening it in 'w' mode blew it away
        with open(self.copy_file("test0.fits"), "wb+") as f:
            with fits.HDUList.fromfile(f) as h:
                # wb+ still causes an existing file to be overwritten so there
                # are no HDUs
                assert len(h) == 0

        # Need to re-copy the file since opening it in 'w' mode blew it away
        testfile = self.copy_file("test0.fits")
        with open(testfile, "rb+") as f:
            with fits.HDUList.fromfile(f) as h:
                assert h.fileinfo(0)["filemode"] == "update"

        with open(testfile, "ab+") as f:
            with fits.HDUList.fromfile(f) as h:
                assert h.fileinfo(0)["filemode"] == "append"

    def test_mmap_unwriteable(self):
        """Regression test for https://github.com/astropy/astropy/issues/968

        Temporarily patches mmap.mmap to exhibit platform-specific bad
        behavior.
        """

        class MockMmap(mmap.mmap):
            def flush(self):
                raise OSError("flush is broken on this platform")

        old_mmap = mmap.mmap
        mmap.mmap = MockMmap

        # Force the mmap test to be rerun
        _File.__dict__["_mmap_available"]._cache.clear()

        try:
            testfile = self.copy_file("test0.fits")
            with pytest.warns(
                AstropyUserWarning, match=r"mmap\.flush is unavailable"
            ) as w:
                with fits.open(testfile, mode="update", memmap=True) as h:
                    h[1].data[0, 0] = 999

            assert len(w) == 1

            # Double check that writing without mmap still worked
            with fits.open(testfile) as h:
                assert h[1].data[0, 0] == 999
        finally:
            mmap.mmap = old_mmap
            _File.__dict__["_mmap_available"]._cache.clear()

    def test_mmap_allocate_error(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/1380

        Temporarily patches mmap.mmap to raise an OSError if mode is ACCESS_COPY.
        """

        mmap_original = mmap.mmap

        # We patch mmap here to raise an error if access=mmap.ACCESS_COPY, which
        # emulates an issue that an OSError is raised if the available address
        # space is less than the size of the file even if memory mapping is used.

        def mmap_patched(*args, **kwargs):
            if kwargs.get("access") == mmap.ACCESS_COPY:
                exc = OSError()
                if sys.platform.startswith("win32"):
                    exc.errno = errno.EINVAL
                    exc.winerror = 1455
                else:
                    exc.errno = errno.ENOMEM
                raise exc
            else:
                return mmap_original(*args, **kwargs)

        with fits.open(self.data("test0.fits"), memmap=True) as hdulist:
            with patch.object(mmap, "mmap", side_effect=mmap_patched) as p:
                with pytest.warns(
                    AstropyUserWarning,
                    match=r"Could not memory map array with mode='readonly'",
                ):
                    data = hdulist[1].data
                p.reset_mock()
            assert not data.flags.writeable

    def test_mmap_closing(self):
        """
        Tests that the mmap reference is closed/removed when there aren't any
        HDU data references left.
        """

        if not _File._mmap_available:
            pytest.xfail("not expected to work on platforms without mmap support")

        with fits.open(self.data("test0.fits"), memmap=True) as hdul:
            assert hdul._file._mmap is None

            hdul[1].data
            assert hdul._file._mmap is not None

            del hdul[1].data
            # Should be no more references to data in the file so close the
            # mmap
            assert hdul._file._mmap is None

            hdul[1].data
            hdul[2].data
            del hdul[1].data
            # hdul[2].data is still references so keep the mmap open
            assert hdul._file._mmap is not None
            del hdul[2].data
            assert hdul._file._mmap is None

        assert hdul._file._mmap is None

        with fits.open(self.data("test0.fits"), memmap=True) as hdul:
            hdul[1].data

        # When the only reference to the data is on the hdu object, and the
        # hdulist it belongs to has been closed, the mmap should be closed as
        # well
        assert hdul._file._mmap is None

        with fits.open(self.data("test0.fits"), memmap=True) as hdul:
            data = hdul[1].data
            # also make a copy
            data_copy = data.copy()

        # The HDUList is closed; in fact, get rid of it completely
        del hdul

        # The data array should still work though...
        assert np.all(data == data_copy)

    def test_uncloseable_file(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2356

        Demonstrates that FITS files can still be read from file-like objects
        that don't have an obvious "open" or "closed" state.
        """

        class MyFileLike:
            def __init__(self, foobar):
                self._foobar = foobar

            def read(self, n):
                return self._foobar.read(n)

            def seek(self, offset, whence=os.SEEK_SET):
                self._foobar.seek(offset, whence)

            def tell(self):
                return self._foobar.tell()

        with open(self.data("test0.fits"), "rb") as f:
            fileobj = MyFileLike(f)

            with fits.open(fileobj) as hdul1:
                with fits.open(self.data("test0.fits")) as hdul2:
                    assert hdul1.info(output=False) == hdul2.info(output=False)
                    for hdu1, hdu2 in zip(hdul1, hdul2):
                        assert hdu1.header == hdu2.header
                        if hdu1.data is not None and hdu2.data is not None:
                            assert np.all(hdu1.data == hdu2.data)

    def test_write_bytesio_discontiguous(self):
        """
        Regression test related to
        https://github.com/astropy/astropy/issues/2794#issuecomment-55441539

        Demonstrates that writing an HDU containing a discontiguous Numpy array
        should work properly.
        """

        data = np.arange(100)[::3]
        hdu = fits.PrimaryHDU(data=data)
        fileobj = io.BytesIO()
        hdu.writeto(fileobj)

        fileobj.seek(0)

        with fits.open(fileobj) as h:
            assert np.all(h[0].data == data)

    def test_write_bytesio(self):
        """
        Regression test for https://github.com/astropy/astropy/issues/2463

        Test against `io.BytesIO`.  `io.StringIO` is not supported.
        """

        self._test_write_string_bytes_io(io.BytesIO())

    @pytest.mark.skipif(
        sys.platform.startswith("win32"), reason="Cannot test on Windows"
    )
    def test_filename_with_colon(self):
        """
        Test reading and writing a file with a colon in the filename.

        Regression test for https://github.com/astropy/astropy/issues/3122
        """

        # Skip on Windows since colons in filenames makes NTFS sad.

        filename = "APEXHET.2014-04-01T15:18:01.000.fits"
        hdu = fits.PrimaryHDU(data=np.arange(10))
        hdu.writeto(self.temp(filename))

        with fits.open(self.temp(filename)) as hdul:
            assert np.all(hdul[0].data == hdu.data)

    def test_writeto_full_disk(self, monkeypatch):
        """
        Test that it gives a readable error when trying to write an hdulist
        to a full disk.
        """

        def _writeto(self, array):
            raise OSError("Fake error raised when writing file.")

        def get_free_space_in_dir(path):
            return 0

        msg = (
            "Not enough space on disk: requested 8000, available 0. "
            "Fake error raised when writing file."
        )
        with pytest.raises(OSError, match=msg) as exc:
            monkeypatch.setattr(fits.hdu.base._BaseHDU, "_writeto", _writeto)
            monkeypatch.setattr(data, "get_free_space_in_dir", get_free_space_in_dir)

            n = np.arange(0, 1000, dtype="int64")
            hdu = fits.PrimaryHDU(n)
            hdulist = fits.HDUList(hdu)
            filename = self.temp("test.fits")

            with open(filename, mode="wb") as fileobj:
                hdulist.writeto(fileobj)

    def test_flush_full_disk(self, monkeypatch):
        """
        Test that it gives a readable error when trying to update an hdulist
        to a full disk.
        """
        filename = self.temp("test.fits")
        hdul = [fits.PrimaryHDU(), fits.ImageHDU()]
        hdul = fits.HDUList(hdul)
        hdul[0].data = np.arange(0, 1000, dtype="int64")
        hdul.writeto(filename)

        def _writedata(self, fileobj):
            raise OSError("Fake error raised when writing file.")

        def get_free_space_in_dir(path):
            return 0

        monkeypatch.setattr(fits.hdu.base._BaseHDU, "_writedata", _writedata)
        monkeypatch.setattr(data, "get_free_space_in_dir", get_free_space_in_dir)

        msg = (
            "Not enough space on disk: requested 8000, available 0. "
            "Fake error raised when writing file."
        )
        with pytest.raises(OSError, match=msg) as exc:
            with fits.open(filename, mode="update") as hdul:
                hdul[0].data = np.arange(0, 1000, dtype="int64")
                hdul.insert(1, fits.ImageHDU())
                hdul.flush()

    def _test_write_string_bytes_io(self, fileobj):
        """
        Implemented for both test_write_stringio and test_write_bytesio.
        """

        with fits.open(self.data("test0.fits")) as hdul:
            hdul.writeto(fileobj)
            hdul2 = fits.HDUList.fromstring(fileobj.getvalue())
            assert FITSDiff(hdul, hdul2).identical

    def _make_gzip_file(self, filename="test0.fits.gz"):
        gzfile = self.temp(filename)
        with open(self.data("test0.fits"), "rb") as f:
            gz = gzip.open(gzfile, "wb")
            gz.write(f.read())
            gz.close()

        return gzfile

    def test_write_overwrite(self, home_is_temp):
        filename = self.temp("test_overwrite.fits")
        hdu = fits.PrimaryHDU(data=np.arange(10))
        hdu.writeto(filename)
        with pytest.raises(OSError, match=_NOT_OVERWRITING_MSG_MATCH):
            hdu.writeto(filename)
        hdu.writeto(filename, overwrite=True)

    def _make_zip_file(self, mode="copyonwrite", filename="test0.fits.zip"):
        zfile = zipfile.ZipFile(self.temp(filename), "w")
        zfile.write(self.data("test0.fits"))
        zfile.close()

        return zfile.filename

    def _make_bzip2_file(self, filename="test0.fits.bz2"):
        bzfile = self.temp(filename)
        with open(self.data("test0.fits"), "rb") as f:
            bz = bz2.BZ2File(bzfile, "w")
            bz.write(f.read())
            bz.close()

        return bzfile

    def _make_lzma_file(self, filename="test0.fits.xz"):
        lzmafile = self.temp(filename)
        with open(self.data("test0.fits"), "rb") as f:
            lz = lzma.LZMAFile(lzmafile, "w")
            lz.write(f.read())
            lz.close()

        return lzmafile

    def _make_lzw_file(self, new_filename=None):
        return self.copy_file("lzw.fits.Z", new_filename)

    def test_simulateonly(self):
        """Write to None simulates writing."""

        with fits.open(self.data("test0.fits")) as hdul:
            hdul.writeto(None)
            hdul[0].writeto(None)
            hdul[0].header.tofile(None)

    def test_bintablehdu_zero_bytes(self):
        """Make sure we don't have any zero-byte writes in BinTableHDU"""

        bright = np.rec.array(
            [
                (1, "Sirius", -1.45, "A1V"),
                (2, "Canopus", -0.73, "F0Ib"),
                (3, "Rigil Kent", -0.1, "G2V"),
            ],
            formats="int16,S20,float32,S10",
            names="order,name,mag,Sp",
        )

        hdu_non_zero = fits.BinTableHDU(bright)
        # use safeio, a special file handler meant to fail on zero-byte writes
        fh = safeio.CatchZeroByteWriter(open(self.temp("bright.fits"), mode="wb"))
        hdu_non_zero.writeto(fh)
        fh.close()

    def test_primaryhdu_zero_bytes(self):
        """
        Make sure we don't have any zero-byte writes from an ImageHDU
        (or other) of `size % BLOCK_SIZE == 0`
        """

        hdu_img_2880 = fits.PrimaryHDU(data=np.arange(720, dtype="i4"))
        # use safeio, a special file handler meant to fail on zero-byte writes
        fh = safeio.CatchZeroByteWriter(open(self.temp("image.fits"), mode="wb"))
        hdu_img_2880.writeto(fh)
        fh.close()

    def test_HDUList_writeto_stdout(self):
        # see https://github.com/astropy/astropy/issues/3427
        hdul = fits.HDUList([fits.PrimaryHDU()])
        hdul.writeto(sys.stdout)


class TestStreamingFunctions(FitsTestCase):
    """Test functionality of the StreamingHDU class."""

    def test_streaming_hdu(self, home_is_temp):
        shdu = self._make_streaming_hdu(self.temp("new.fits"))
        assert isinstance(shdu.size, int)
        assert shdu.size == 100

        arr = np.arange(25, dtype=np.int32).reshape((5, 5))
        shdu.write(arr)
        assert shdu.writecomplete
        shdu.close()

        with fits.open(self.temp("new.fits")) as hdul:
            assert len(hdul) == 1
            assert (hdul[0].data == arr).all()

    def test_streaming_hdu_file_wrong_mode(self):
        """
        Test that streaming an HDU to a file opened in the wrong mode fails as
        expected.
        """
        with pytest.raises(ValueError):
            with open(self.temp("new.fits"), "wb") as f:
                header = fits.Header()
                fits.StreamingHDU(f, header)

    def test_streaming_hdu_write_file(self):
        """Test streaming an HDU to an open file object."""

        arr = np.zeros((5, 5), dtype=np.int32)
        with open(self.temp("new.fits"), "ab+") as f:
            shdu = self._make_streaming_hdu(f)
            shdu.write(arr)
            assert shdu.writecomplete
            assert shdu.size == 100
        with fits.open(self.temp("new.fits")) as hdul:
            assert len(hdul) == 1
            assert (hdul[0].data == arr).all()

    def test_streaming_hdu_write_file_like(self):
        """Test streaming an HDU to an open file-like object."""

        arr = np.zeros((5, 5), dtype=np.int32)
        # The file-like object underlying a StreamingHDU must be in binary mode
        sf = io.BytesIO()
        shdu = self._make_streaming_hdu(sf)
        shdu.write(arr)
        assert shdu.writecomplete
        assert shdu.size == 100

        sf.seek(0)
        hdul = fits.open(sf)
        assert len(hdul) == 1
        assert (hdul[0].data == arr).all()

    def test_streaming_hdu_append_extension(self):
        arr = np.zeros((5, 5), dtype=np.int32)
        with open(self.temp("new.fits"), "ab+") as f:
            shdu = self._make_streaming_hdu(f)
            shdu.write(arr)
        # Doing this again should update the file with an extension
        with open(self.temp("new.fits"), "ab+") as f:
            shdu = self._make_streaming_hdu(f)
            shdu.write(arr)

    def test_fix_invalid_extname(self, capsys):
        phdu = fits.PrimaryHDU()
        ihdu = fits.ImageHDU()
        ihdu.header["EXTNAME"] = 12345678
        hdul = fits.HDUList([phdu, ihdu])
        filename = self.temp("temp.fits")

        with pytest.raises(fits.VerifyError):
            hdul.writeto(filename, output_verify="exception")
        with pytest.warns(fits.verify.VerifyWarning) as w:
            hdul.writeto(filename, output_verify="fix")
        assert "Verification reported errors" in str(w[0].message)
        with fits.open(filename):
            assert hdul[1].name == "12345678"
            assert hdul[1].header["EXTNAME"] == "12345678"

        hdul.close()

    def _make_streaming_hdu(self, fileobj):
        hd = fits.Header()
        hd["SIMPLE"] = (True, "conforms to FITS standard")
        hd["BITPIX"] = (32, "array data type")
        hd["NAXIS"] = (2, "number of array dimensions")
        hd["NAXIS1"] = 5
        hd["NAXIS2"] = 5
        hd["EXTEND"] = True
        return fits.StreamingHDU(fileobj, hd)

    def test_blank_ignore(self):
        with fits.open(self.data("blank.fits"), ignore_blank=True) as f:
            assert f[0].data.item(0) == 2

    def test_error_if_memmap_impossible(self):
        pth = self.data("blank.fits")
        with fits.open(pth, memmap=True) as hdul:
            with pytest.raises(ValueError):
                hdul[0].data

        # However, it should not fail if do_not_scale_image_data was used:
        # See https://github.com/astropy/astropy/issues/3766
        with fits.open(pth, memmap=True, do_not_scale_image_data=True) as hdul:
            hdul[0].data  # Just make sure it doesn't crash



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Cosmological units."""

__all__ = ["littleh", "redshift"]

from typing import Final

import astropy.units as u

_ns = globals()


# This is not formally a unit, but is used in that way in many contexts, and
# an appropriate equivalency is only possible if it's treated as a unit.
redshift: Final = u.def_unit(
    ["redshift"],
    prefixes=False,
    namespace=_ns,
    doc="Cosmological redshift.",
    format={"latex": r""},
)
u.def_physical_type(redshift, "redshift")

# This is not formally a unit, but is used in that way in many contexts, and
# an appropriate equivalency is only possible if it's treated as a unit (see
# https://arxiv.org/pdf/1308.4150.pdf for more)
# Also note that h or h100 or h_100 would be a better name, but they either
# conflict or have numbers in them, which is disallowed
littleh: Final = u.def_unit(
    ["littleh"],
    namespace=_ns,
    prefixes=False,
    doc='Reduced/"dimensionless" Hubble constant',
    format={"latex": r"h_{100}"},
)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import json
import locale
import os
import urllib.error
from datetime import datetime

import numpy as np
import pytest

from astropy.io import fits
from astropy.tests.helper import CI
from astropy.utils import data, misc
from astropy.utils.exceptions import AstropyDeprecationWarning


@pytest.mark.parametrize(
    "obj,expectation",
    [
        (2, False),
        ([2], True),
        ([1, 2, 3], True),
        (np.array(2), False),
        (np.array([1, 2, 3]), True),
    ],
)
def test_isiterable(obj, expectation):
    with pytest.warns(
        AstropyDeprecationWarning, match=r"Use numpy.iterable\(\) instead\.$"
    ):
        assert misc.isiterable(obj) is expectation


@pytest.mark.remote_data
def test_api_lookup():
    try:
        strurl = misc.find_api_page("astropy.utils.misc", "dev", False, timeout=5)
        objurl = misc.find_api_page(misc, "dev", False, timeout=5)
    except (urllib.error.URLError, TimeoutError):
        if CI:
            pytest.xfail("Timed out in CI")
        else:
            raise

    assert strurl == objurl
    assert (
        strurl
        == "http://devdocs.astropy.org/utils/ref_api.html#module-astropy.utils.misc"
    )

    # Try a non-dev version
    objurl = misc.find_api_page(misc, "stable", False, timeout=3)
    assert (
        objurl
        == "https://docs.astropy.org/en/stable/utils/ref_api.html#module-astropy.utils.misc"
    )


def test_is_path_hidden_deprecation():
    with pytest.warns(
        AstropyDeprecationWarning, match="^The is_path_hidden function is deprecated"
    ):
        misc.is_path_hidden("data")


# This is the only test that uses astropy/utils/tests/data/.hidden_file.txt
def test_skip_hidden():
    path = data.get_pkg_data_path("data")
    for _, _, files in os.walk(path):
        assert ".hidden_file.txt" in files
        assert "local.dat" in files
        # break after the first level since the data dir contains some other
        # subdirectories that don't have these files
        break
    with pytest.warns(
        AstropyDeprecationWarning, match="^The .*_hidden function is deprecated"
    ):
        for _, _, files in misc.walk_skip_hidden(path):
            assert ".hidden_file.txt" not in files
            assert "local.dat" in files
            break


def test_JsonCustomEncoder():
    from astropy import units as u

    assert json.dumps(np.arange(3), cls=misc.JsonCustomEncoder) == "[0, 1, 2]"
    assert json.dumps(1 + 2j, cls=misc.JsonCustomEncoder) == "[1.0, 2.0]"
    assert json.dumps({1, 2}, cls=misc.JsonCustomEncoder) == "[1, 2]"
    assert (
        json.dumps(b"hello world \xc3\x85", cls=misc.JsonCustomEncoder)
        == '"hello world \\u00c5"'
    )
    assert json.dumps({1: 2}, cls=misc.JsonCustomEncoder) == '{"1": 2}'  # default
    assert json.dumps({1: u.m}, cls=misc.JsonCustomEncoder) == '{"1": "m"}'
    # Quantities
    tmp = json.dumps({"a": 5 * u.cm}, cls=misc.JsonCustomEncoder)
    newd = json.loads(tmp)
    tmpd = {"a": {"unit": "cm", "value": 5.0}}
    assert newd == tmpd
    tmp2 = json.dumps({"a": np.arange(2) * u.cm}, cls=misc.JsonCustomEncoder)
    newd = json.loads(tmp2)
    tmpd = {"a": {"unit": "cm", "value": [0.0, 1.0]}}
    assert newd == tmpd
    tmp3 = json.dumps({"a": np.arange(2) * u.erg / u.s}, cls=misc.JsonCustomEncoder)
    newd = json.loads(tmp3)
    tmpd = {"a": {"unit": "erg / s", "value": [0.0, 1.0]}}
    assert newd == tmpd


def test_JsonCustomEncoder_FITS_rec_from_files():
    with fits.open(
        fits.util.get_testdata_filepath("variable_length_table.fits")
    ) as hdul:
        assert (
            json.dumps(hdul[1].data, cls=misc.JsonCustomEncoder)
            == "[[[45, 56], [11, 3]], [[11, 12, 13], [12, 4]]]"
        )

    with fits.open(fits.util.get_testdata_filepath("btable.fits")) as hdul:
        assert (
            json.dumps(hdul[1].data, cls=misc.JsonCustomEncoder)
            == '[[1, "Sirius", -1.4500000476837158, "A1V"], '
            '[2, "Canopus", -0.7300000190734863, "F0Ib"], '
            '[3, "Rigil Kent", -0.10000000149011612, "G2V"]]'
        )

    with fits.open(fits.util.get_testdata_filepath("table.fits")) as hdul:
        assert (
            json.dumps(hdul[1].data, cls=misc.JsonCustomEncoder)
            == '[["NGC1001", 11.100000381469727], '
            '["NGC1002", 12.300000190734863], '
            '["NGC1003", 15.199999809265137]]'
        )


def test_set_locale():
    # First, test if the required locales are available
    current = locale.setlocale(locale.LC_ALL)
    try:
        locale.setlocale(locale.LC_ALL, "en_US.utf8")
        locale.setlocale(locale.LC_ALL, "fr_FR.utf8")
    except locale.Error as e:
        pytest.skip(f"Locale error: {e}")
    finally:
        locale.setlocale(locale.LC_ALL, current)

    date = datetime(2000, 10, 1, 0, 0, 0)
    day_mon = date.strftime("%a, %b")

    with misc._set_locale("en_US.utf8"):
        assert date.strftime("%a, %b") == "Sun, Oct"

    with misc._set_locale("fr_FR.utf8"):
        assert date.strftime("%a, %b") == "dim., oct."

    # Back to original
    assert date.strftime("%a, %b") == day_mon

    with misc._set_locale(current):
        assert date.strftime("%a, %b") == day_mon


def test_dtype_bytes_or_chars():
    assert misc.dtype_bytes_or_chars(np.dtype(np.float64)) == 8
    assert misc.dtype_bytes_or_chars(np.dtype(object)) is None
    assert misc.dtype_bytes_or_chars(np.dtype(np.int32)) == 4
    assert misc.dtype_bytes_or_chars(np.array(b"12345").dtype) == 5
    assert misc.dtype_bytes_or_chars(np.array("12345").dtype) == 5


def test_indent_deprecation():
    with pytest.warns(AstropyDeprecationWarning, match=r"Use textwrap\.indent"):
        misc.indent("Obsolete since Python 3.3")


def test_format_exception_deprecation():
    with pytest.warns(AstropyDeprecationWarning):
        misc.format_exception("this is deprecated")



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module tests some of the methods related to the ``HTML``
reader/writer and aims to document its functionality.

Requires `BeautifulSoup <http://www.crummy.com/software/BeautifulSoup/>`_
to be installed.
"""

import os
from io import StringIO
from pathlib import Path

import numpy as np
import pytest

from astropy.io import ascii
from astropy.io.ascii import core, html
from astropy.table import Table
from astropy.utils.compat.optional_deps import HAS_BLEACH, HAS_BS4

from .common import setup_function, teardown_function  # noqa: F401

if HAS_BS4:
    from bs4 import BeautifulSoup, FeatureNotFound


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_soupstring():
    """
    Test to make sure the class SoupString behaves properly.
    """

    soup = BeautifulSoup(
        "<html><head></head><body><p>foo</p></body></html>", "html.parser"
    )
    soup_str = html.SoupString(soup)
    assert isinstance(soup_str, str)
    assert isinstance(soup_str, html.SoupString)
    assert soup_str == "<html><head></head><body><p>foo</p></body></html>"
    assert soup_str.soup is soup


def test_listwriter():
    """
    Test to make sure the class ListWriter behaves properly.
    """

    lst = []
    writer = html.ListWriter(lst)

    for i in range(5):
        writer.write(i)
    for ch in "abcde":
        writer.write(ch)

    assert lst == [0, 1, 2, 3, 4, "a", "b", "c", "d", "e"]


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_identify_table():
    """
    Test to make sure that identify_table() returns whether the
    given BeautifulSoup tag is the correct table to process.
    """

    # Should return False on non-<table> tags and None
    soup = BeautifulSoup("<html><body></body></html>", "html.parser")
    assert html.identify_table(soup, {}, 0) is False
    assert html.identify_table(None, {}, 0) is False

    soup = BeautifulSoup(
        '<table id="foo"><tr><th>A</th></tr><tr><td>B</td></tr></table>',
        "html.parser",
    ).table
    assert html.identify_table(soup, {}, 2) is False
    assert html.identify_table(soup, {}, 1) is True  # Default index of 1

    # Same tests, but with explicit parameter
    assert html.identify_table(soup, {"table_id": 2}, 1) is False
    assert html.identify_table(soup, {"table_id": 1}, 1) is True

    # Test identification by string ID
    assert html.identify_table(soup, {"table_id": "bar"}, 1) is False
    assert html.identify_table(soup, {"table_id": "foo"}, 1) is True


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_missing_data():
    """
    Test reading a table with missing data
    """
    # First with default where blank => '0'
    table_in = [
        "<table>",
        "<tr><th>A</th></tr>",
        "<tr><td></td></tr>",
        "<tr><td>1</td></tr>",
        "</table>",
    ]
    dat = Table.read(table_in, format="ascii.html")
    assert dat.masked is False
    assert np.all(dat["A"].mask == [True, False])
    assert dat["A"].dtype.kind == "i"

    # Now with a specific value '...' => missing
    table_in = [
        "<table>",
        "<tr><th>A</th></tr>",
        "<tr><td>...</td></tr>",
        "<tr><td>1</td></tr>",
        "</table>",
    ]
    dat = Table.read(table_in, format="ascii.html", fill_values=[("...", "0")])
    assert dat.masked is False
    assert np.all(dat["A"].mask == [True, False])
    assert dat["A"].dtype.kind == "i"


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_rename_cols():
    """
    Test reading a table and renaming cols
    """
    table_in = [
        "<table>",
        "<tr><th>A</th> <th>B</th></tr>",
        "<tr><td>1</td><td>2</td></tr>",
        "</table>",
    ]

    # Swap column names
    dat = Table.read(table_in, format="ascii.html", names=["B", "A"])
    assert dat.colnames == ["B", "A"]
    assert len(dat) == 1

    # Swap column names and only include A (the renamed version)
    dat = Table.read(
        table_in, format="ascii.html", names=["B", "A"], include_names=["A"]
    )
    assert dat.colnames == ["A"]
    assert len(dat) == 1
    assert np.all(dat["A"] == 2)


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_no_names():
    """
    Test reading a table with no column header
    """
    table_in = ["<table>", "<tr><td>1</td></tr>", "<tr><td>2</td></tr>", "</table>"]
    dat = Table.read(table_in, format="ascii.html")
    assert dat.colnames == ["col1"]
    assert len(dat) == 2

    dat = Table.read(table_in, format="ascii.html", names=["a"])
    assert dat.colnames == ["a"]
    assert len(dat) == 2


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_identify_table_fail():
    """
    Raise an exception with an informative error message if table_id
    is not found.
    """
    table_in = ['<table id="foo"><tr><th>A</th></tr>', "<tr><td>B</td></tr></table>"]

    with pytest.raises(core.InconsistentTableError) as err:
        Table.read(
            table_in, format="ascii.html", htmldict={"table_id": "bad_id"}, guess=False
        )
    assert err.match("ERROR: HTML table id 'bad_id' not found$")

    with pytest.raises(core.InconsistentTableError) as err:
        Table.read(table_in, format="ascii.html", htmldict={"table_id": 3}, guess=False)
    assert err.match("ERROR: HTML table number 3 not found$")


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_backend_parsers():
    """
    Make sure the user can specify which back-end parser to use
    and that an error is raised if the parser is invalid.
    """
    for parser in ("lxml", "xml", "html.parser", "html5lib"):
        try:
            Table.read(
                "data/html2.html",
                format="ascii.html",
                htmldict={"parser": parser},
                guess=False,
            )
        except FeatureNotFound:
            if parser == "html.parser":
                raise
            # otherwise ignore if the dependency isn't present

    # reading should fail if the parser is invalid
    with pytest.raises(FeatureNotFound):
        Table.read(
            "data/html2.html",
            format="ascii.html",
            htmldict={"parser": "foo"},
            guess=False,
        )


@pytest.mark.skipif(HAS_BS4, reason="requires no BeautifulSoup4")
def test_htmlinputter_no_bs4():
    """
    This should return an OptionalTableImportError if BeautifulSoup
    is not installed.
    """

    inputter = html.HTMLInputter()
    with pytest.raises(core.OptionalTableImportError):
        inputter.process_lines([])


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_htmlinputter():
    """
    Test to ensure that HTMLInputter correctly converts input
    into a list of SoupStrings representing table elements.
    """

    f = "data/html.html"
    with open(f) as fd:
        table = fd.read()

    inputter = html.HTMLInputter()
    inputter.html = {}

    # In absence of table_id, defaults to the first table
    expected = [
        "<tr><th>Column 1</th><th>Column 2</th><th>Column 3</th></tr>",
        "<tr><td>1</td><td>a</td><td>1.05</td></tr>",
        "<tr><td>2</td><td>b</td><td>2.75</td></tr>",
        "<tr><td>3</td><td>c</td><td>-1.25</td></tr>",
    ]
    assert [str(x) for x in inputter.get_lines(table)] == expected

    # Should raise an InconsistentTableError if the table is not found
    inputter.html = {"table_id": 4}
    with pytest.raises(core.InconsistentTableError):
        inputter.get_lines(table)

    # Identification by string ID
    inputter.html["table_id"] = "second"
    expected = [
        "<tr><th>Column A</th><th>Column B</th><th>Column C</th></tr>",
        "<tr><td>4</td><td>d</td><td>10.5</td></tr>",
        "<tr><td>5</td><td>e</td><td>27.5</td></tr>",
        "<tr><td>6</td><td>f</td><td>-12.5</td></tr>",
    ]
    assert [str(x) for x in inputter.get_lines(table)] == expected

    # Identification by integer index
    inputter.html["table_id"] = 3
    expected = [
        "<tr><th>C1</th><th>C2</th><th>C3</th></tr>",
        "<tr><td>7</td><td>g</td><td>105.0</td></tr>",
        "<tr><td>8</td><td>h</td><td>275.0</td></tr>",
        "<tr><td>9</td><td>i</td><td>-125.0</td></tr>",
    ]
    assert [str(x) for x in inputter.get_lines(table)] == expected


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_htmlsplitter():
    """
    Test to make sure that HTMLSplitter correctly inputs lines
    of type SoupString to return a generator that gives all
    header and data elements.
    """

    splitter = html.HTMLSplitter()

    lines = [
        html.SoupString(
            BeautifulSoup(
                "<table><tr><th>Col 1</th><th>Col 2</th></tr></table>", "html.parser"
            ).tr
        ),
        html.SoupString(
            BeautifulSoup(
                "<table><tr><td>Data 1</td><td>Data 2</td></tr></table>", "html.parser"
            ).tr
        ),
    ]
    expected_data = [["Col 1", "Col 2"], ["Data 1", "Data 2"]]
    assert list(splitter(lines)) == expected_data

    # Make sure the presence of a non-SoupString triggers a TypeError
    lines.append("<tr><td>Data 3</td><td>Data 4</td></tr>")
    with pytest.raises(TypeError):
        list(splitter(lines))

    # Make sure that passing an empty list triggers an error
    with pytest.raises(core.InconsistentTableError):
        list(splitter([]))


@pytest.mark.parametrize(
    "get_table",
    [
        lambda path: os.fspath(path),
        lambda path: Path(path),
        lambda path: Path(path).read_text(),
    ],
)
@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_htmlheader_start(get_table):
    """
    Test to ensure that the start_line method of HTMLHeader
    returns the first line of header data. Uses t/html.html
    for sample input.
    """

    table_file = "data/html.html"
    table = get_table(table_file)

    inputter = html.HTMLInputter()
    inputter.html = {}
    header = html.HTMLHeader()

    lines = inputter.get_lines(table)
    assert (
        str(lines[header.start_line(lines)])
        == "<tr><th>Column 1</th><th>Column 2</th><th>Column 3</th></tr>"
    )
    inputter.html["table_id"] = "second"
    lines = inputter.get_lines(table)
    assert (
        str(lines[header.start_line(lines)])
        == "<tr><th>Column A</th><th>Column B</th><th>Column C</th></tr>"
    )
    inputter.html["table_id"] = 3
    lines = inputter.get_lines(table)
    assert (
        str(lines[header.start_line(lines)])
        == "<tr><th>C1</th><th>C2</th><th>C3</th></tr>"
    )

    # start_line should return None if no valid header is found
    lines = [
        html.SoupString(
            BeautifulSoup("<table><tr><td>Data</td></tr></table>", "html.parser").tr
        ),
        html.SoupString(BeautifulSoup("<p>Text</p>", "html.parser").p),
    ]
    assert header.start_line(lines) is None

    # Should raise an error if a non-SoupString is present
    lines.append("<tr><th>Header</th></tr>")
    with pytest.raises(TypeError):
        header.start_line(lines)


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_htmldata():
    """
    Test to ensure that the start_line and end_lines methods
    of HTMLData returns the first line of table data. Uses
    t/html.html for sample input.
    """

    f = "data/html.html"
    with open(f) as fd:
        table = fd.read()

    inputter = html.HTMLInputter()
    inputter.html = {}
    data = html.HTMLData()

    lines = inputter.get_lines(table)
    assert (
        str(lines[data.start_line(lines)])
        == "<tr><td>1</td><td>a</td><td>1.05</td></tr>"
    )
    # end_line returns the index of the last data element + 1
    assert (
        str(lines[data.end_line(lines) - 1])
        == "<tr><td>3</td><td>c</td><td>-1.25</td></tr>"
    )

    inputter.html["table_id"] = "second"
    lines = inputter.get_lines(table)
    assert (
        str(lines[data.start_line(lines)])
        == "<tr><td>4</td><td>d</td><td>10.5</td></tr>"
    )
    assert (
        str(lines[data.end_line(lines) - 1])
        == "<tr><td>6</td><td>f</td><td>-12.5</td></tr>"
    )

    inputter.html["table_id"] = 3
    lines = inputter.get_lines(table)
    assert (
        str(lines[data.start_line(lines)])
        == "<tr><td>7</td><td>g</td><td>105.0</td></tr>"
    )
    assert (
        str(lines[data.end_line(lines) - 1])
        == "<tr><td>9</td><td>i</td><td>-125.0</td></tr>"
    )

    # start_line should raise an error if no table data exists
    lines = [
        html.SoupString(BeautifulSoup("<div></div>", "html.parser").div),
        html.SoupString(BeautifulSoup("<p>Text</p>", "html.parser").p),
    ]
    with pytest.raises(core.InconsistentTableError):
        data.start_line(lines)

    # end_line should return None if no table data exists
    assert data.end_line(lines) is None

    # Should raise an error if a non-SoupString is present
    lines.append("<tr><td>Data</td></tr>")
    with pytest.raises(TypeError):
        data.start_line(lines)
    with pytest.raises(TypeError):
        data.end_line(lines)


def test_multicolumn_write():
    """
    Test to make sure that the HTML writer writes multidimensional
    columns (those with iterable elements) using the colspan
    attribute of <th>.
    """

    col1 = [1, 2, 3]
    col2 = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    col3 = [("a", "a", "a"), ("b", "b", "b"), ("c", "c", "c")]
    table = Table([col1, col2, col3], names=("C1", "C2", "C3"))
    expected = """\
<html>
 <head>
  <meta charset="utf-8"/>
  <meta content="text/html;charset=UTF-8" http-equiv="Content-type"/>
 </head>
 <body>
  <table>
   <thead>
    <tr>
     <th>C1</th>
     <th colspan="2">C2</th>
     <th colspan="3">C3</th>
    </tr>
   </thead>
   <tr>
    <td>1</td>
    <td>1.0</td>
    <td>1.0</td>
    <td>a</td>
    <td>a</td>
    <td>a</td>
   </tr>
   <tr>
    <td>2</td>
    <td>2.0</td>
    <td>2.0</td>
    <td>b</td>
    <td>b</td>
    <td>b</td>
   </tr>
   <tr>
    <td>3</td>
    <td>3.0</td>
    <td>3.0</td>
    <td>c</td>
    <td>c</td>
    <td>c</td>
   </tr>
  </table>
 </body>
</html>
    """
    out = html.HTML().write(table)[0].strip()
    assert out == expected.strip()


@pytest.mark.skipif(not HAS_BLEACH, reason="requires bleach")
def test_multicolumn_write_escape():
    """
    Test to make sure that the HTML writer writes multidimensional
    columns (those with iterable elements) using the colspan
    attribute of <th>.
    """

    col1 = [1, 2, 3]
    col2 = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    col3 = [("<a></a>", "<a></a>", "a"), ("<b></b>", "b", "b"), ("c", "c", "c")]
    table = Table([col1, col2, col3], names=("C1", "C2", "C3"))
    expected = """\
<html>
 <head>
  <meta charset="utf-8"/>
  <meta content="text/html;charset=UTF-8" http-equiv="Content-type"/>
 </head>
 <body>
  <table>
   <thead>
    <tr>
     <th>C1</th>
     <th colspan="2">C2</th>
     <th colspan="3">C3</th>
    </tr>
   </thead>
   <tr>
    <td>1</td>
    <td>1.0</td>
    <td>1.0</td>
    <td><a></a></td>
    <td><a></a></td>
    <td>a</td>
   </tr>
   <tr>
    <td>2</td>
    <td>2.0</td>
    <td>2.0</td>
    <td><b></b></td>
    <td>b</td>
    <td>b</td>
   </tr>
   <tr>
    <td>3</td>
    <td>3.0</td>
    <td>3.0</td>
    <td>c</td>
    <td>c</td>
    <td>c</td>
   </tr>
  </table>
 </body>
</html>
    """
    out = html.HTML(htmldict={"raw_html_cols": "C3"}).write(table)[0].strip()
    assert out == expected.strip()


def test_write_no_multicols():
    """
    Test to make sure that the HTML writer will not use
    multi-dimensional columns if the multicol parameter
    is False.
    """

    col1 = [1, 2, 3]
    col2 = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    col3 = [("a", "a", "a"), ("b", "b", "b"), ("c", "c", "c")]
    table = Table([col1, col2, col3], names=("C1", "C2", "C3"))
    expected = """\
<html>
 <head>
  <meta charset="utf-8"/>
  <meta content="text/html;charset=UTF-8" http-equiv="Content-type"/>
 </head>
 <body>
  <table>
   <thead>
    <tr>
     <th>C1</th>
     <th>C2</th>
     <th>C3</th>
    </tr>
   </thead>
   <tr>
    <td>1</td>
    <td>1.0 .. 1.0</td>
    <td>a .. a</td>
   </tr>
   <tr>
    <td>2</td>
    <td>2.0 .. 2.0</td>
    <td>b .. b</td>
   </tr>
   <tr>
    <td>3</td>
    <td>3.0 .. 3.0</td>
    <td>c .. c</td>
   </tr>
  </table>
 </body>
</html>
    """
    assert html.HTML({"multicol": False}).write(table)[0].strip() == expected.strip()


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_multicolumn_read():
    """
    Test to make sure that the HTML reader inputs multidimensional
    columns (those with iterable elements) using the colspan
    attribute of <th>.

    Ensure that any string element within a multidimensional column
    casts all elements to string prior to type conversion operations.
    """

    table = Table.read("data/html2.html", format="ascii.html")
    str_type = np.dtype((str, 21))
    expected = Table(
        np.array(
            [(["1", "2.5000000000000000001"], 3), (["1a", "1"], 3.5)],
            dtype=[("A", str_type, (2,)), ("B", "<f8")],
        )
    )
    assert np.all(table == expected)


@pytest.mark.skipif(not HAS_BLEACH, reason="requires bleach")
def test_raw_html_write():
    """
    Test that columns can contain raw HTML which is not escaped.
    """
    t = Table([["<em>x</em>"], ["<em>y</em>"]], names=["a", "b"])

    # One column contains raw HTML (string input)
    out = StringIO()
    t.write(out, format="ascii.html", htmldict={"raw_html_cols": "a"})
    expected = """\
   <tr>
    <td><em>x</em></td>
    <td>&lt;em&gt;y&lt;/em&gt;</td>
   </tr>"""
    assert expected in out.getvalue()

    # One column contains raw HTML (list input)
    out = StringIO()
    t.write(out, format="ascii.html", htmldict={"raw_html_cols": ["a"]})
    assert expected in out.getvalue()

    # Two columns contains raw HTML (list input)
    out = StringIO()
    t.write(out, format="ascii.html", htmldict={"raw_html_cols": ["a", "b"]})
    expected = """\
   <tr>
    <td><em>x</em></td>
    <td><em>y</em></td>
   </tr>"""
    assert expected in out.getvalue()


@pytest.mark.skipif(not HAS_BLEACH, reason="requires bleach")
def test_raw_html_write_clean():
    """
    Test that columns can contain raw HTML which is not escaped.
    """
    import bleach

    t = Table(
        [["<script>x</script>"], ["<p>y</p>"], ["<em>y</em>"]], names=["a", "b", "c"]
    )

    # Confirm that <script> and <p> get escaped but not <em>
    out = StringIO()
    t.write(out, format="ascii.html", htmldict={"raw_html_cols": t.colnames})
    expected = """\
   <tr>
    <td>&lt;script&gt;x&lt;/script&gt;</td>
    <td>&lt;p&gt;y&lt;/p&gt;</td>
    <td><em>y</em></td>
   </tr>"""
    assert expected in out.getvalue()

    # Confirm that we can whitelist <p>
    out = StringIO()
    t.write(
        out,
        format="ascii.html",
        htmldict={
            "raw_html_cols": t.colnames,
            "raw_html_clean_kwargs": {"tags": list(bleach.ALLOWED_TAGS) + ["p"]},
        },
    )
    expected = """\
   <tr>
    <td>&lt;script&gt;x&lt;/script&gt;</td>
    <td><p>y</p></td>
    <td><em>y</em></td>
   </tr>"""
    assert expected in out.getvalue()


def test_write_table_html_fill_values():
    """
    Test that passing fill_values should replace any matching row
    """
    buffer_output = StringIO()
    t = Table([[1], [2]], names=("a", "b"))
    ascii.write(t, buffer_output, fill_values=("1", "Hello world"), format="html")

    t_expected = Table([["Hello world"], [2]], names=("a", "b"))
    buffer_expected = StringIO()
    ascii.write(t_expected, buffer_expected, format="html")

    assert buffer_output.getvalue() == buffer_expected.getvalue()


def test_write_table_html_fill_values_optional_columns():
    """
    Test that passing optional column in fill_values should only replace
    matching columns
    """
    buffer_output = StringIO()
    t = Table([[1], [1]], names=("a", "b"))
    ascii.write(t, buffer_output, fill_values=("1", "Hello world", "b"), format="html")

    t_expected = Table([[1], ["Hello world"]], names=("a", "b"))
    buffer_expected = StringIO()
    ascii.write(t_expected, buffer_expected, format="html")

    assert buffer_output.getvalue() == buffer_expected.getvalue()


def test_write_table_html_fill_values_masked():
    """
    Test that passing masked values in fill_values should only replace
    masked columns or values
    """
    buffer_output = StringIO()
    t = Table([[1], [1]], names=("a", "b"), masked=True, dtype=("i4", "i8"))
    t["a"] = np.ma.masked
    ascii.write(t, buffer_output, fill_values=(ascii.masked, "TEST"), format="html")

    t_expected = Table([["TEST"], [1]], names=("a", "b"))
    buffer_expected = StringIO()
    ascii.write(t_expected, buffer_expected, format="html")

    assert buffer_output.getvalue() == buffer_expected.getvalue()


def test_multicolumn_table_html_fill_values():
    """
    Test to make sure that the HTML writer writes multidimensional
    columns with correctly replaced fill_values.
    """
    col1 = [1, 2, 3]
    col2 = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    col3 = [("a", "a", "a"), ("b", "b", "b"), ("c", "c", "c")]

    buffer_output = StringIO()
    t = Table([col1, col2, col3], names=("C1", "C2", "C3"))
    ascii.write(t, buffer_output, fill_values=("a", "z"), format="html")

    col1 = [1, 2, 3]
    col2 = [(1.0, 1.0), (2.0, 2.0), (3.0, 3.0)]
    col3 = [("z", "z", "z"), ("b", "b", "b"), ("c", "c", "c")]

    buffer_expected = StringIO()
    t_expected = Table([col1, col2, col3], names=("C1", "C2", "C3"))
    ascii.write(t_expected, buffer_expected, format="html")

    assert buffer_output.getvalue() == buffer_expected.getvalue()


def test_multi_column_write_table_html_fill_values_masked():
    """
    Test that passing masked values in fill_values should only replace
    masked columns or values for multidimensional tables
    """
    buffer_output = StringIO()
    t = Table([[1, 2, 3, 4], ["--", "a", "--", "b"]], names=("a", "b"), masked=True)
    t["a"][0:2] = np.ma.masked
    t["b"][0:2] = np.ma.masked
    ascii.write(t, buffer_output, fill_values=[(ascii.masked, "MASKED")], format="html")

    t_expected = Table(
        [["MASKED", "MASKED", 3, 4], ["MASKED", "MASKED", "--", "b"]], names=("a", "b")
    )
    buffer_expected = StringIO()
    ascii.write(t_expected, buffer_expected, format="html")
    print(buffer_expected.getvalue())

    assert buffer_output.getvalue() == buffer_expected.getvalue()


def test_write_table_formatted_columns():
    """
    Test to make sure that the HTML writer writes out using the
    supplied formatting.
    """

    col1 = [1, 2]
    col2 = [1.234567e-11, -9.876543e11]
    formats = {"C1": "04d", "C2": ".2e"}
    table = Table([col1, col2], names=formats.keys())

    expected = """\
<html>
 <head>
  <meta charset="utf-8"/>
  <meta content="text/html;charset=UTF-8" http-equiv="Content-type"/>
 </head>
 <body>
  <table>
   <thead>
    <tr>
     <th>C1</th>
     <th>C2</th>
    </tr>
   </thead>
   <tr>
    <td>0001</td>
    <td>1.23e-11</td>
   </tr>
   <tr>
    <td>0002</td>
    <td>-9.88e+11</td>
   </tr>
  </table>
 </body>
</html>
    """
    with StringIO() as sp:
        table.write(sp, format="html", formats=formats)
        out = sp.getvalue().strip()
    assert out == expected.strip()


@pytest.mark.skipif(not HAS_BS4, reason="requires BeautifulSoup4")
def test_read_html_unicode():
    """
    Test reading an HTML table with unicode values
    """
    table_in = [
        "<table>",
        "<tr><td>&#x0394;</td></tr>",
        "<tr><td>Δ</td></tr>",
        "</table>",
    ]
    dat = Table.read(table_in, format="ascii.html")
    assert np.all(dat["col1"] == ["Δ", "Δ"])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Handles the "VOUnit" unit format.
"""

import re
import warnings
from re import Pattern
from typing import ClassVar, Literal

import numpy as np

from astropy.extern.ply.lex import LexToken
from astropy.units.core import (
    CompositeUnit,
    NamedUnit,
    PrefixUnit,
    UnitBase,
    def_unit,
    dimensionless_unscaled,
    si_prefixes,
)
from astropy.units.enums import DeprecatedUnitAction
from astropy.units.errors import UnitParserWarning, UnitScaleError, UnitsError
from astropy.units.typing import UnitScale
from astropy.utils import classproperty

from . import Base
from .generic import _GenericParserMixin


class VOUnit(Base, _GenericParserMixin):
    """
    The IVOA standard for units used by the VO.

    This is an implementation of `Units in the VO 1.0
    <https://www.ivoa.net/documents/VOUnits/20140523/index.html>`_.
    """

    _explicit_custom_unit_regex: ClassVar[Pattern[str]] = re.compile(
        r"^[YZEPTGMkhdcmunpfazy]?'((?!\d)\w)+'$"
    )
    _custom_unit_regex: ClassVar[Pattern[str]] = re.compile(r"^((?!\d)\w)+$")
    _custom_units: ClassVar[dict[str, UnitBase]] = {}
    _space: ClassVar[str] = "."
    _scale_unit_separator: ClassVar[str] = ""

    @classproperty(lazy=True)
    def _all_units(cls) -> tuple[dict[str, UnitBase], frozenset[str]]:
        from astropy import units as u
        from astropy.units import required_by_vounit as uvo

        names = {"as": u.attosecond}
        deprecated_names = set()
        # The tropical year is missing here compared to the standard
        bases = [
            "A", "a", "adu", "arcmin", "arcsec", "barn", "beam", "bin",
            "C", "cd", "chan", "count", "ct", "d", "D", "deg", "erg", "eV",
            "F", "g", "G", "H", "h", "Hz", "J", "Jy", "K", "lm", "lx", "lyr",
            "m", "mag", "min", "mol", "N", "Ohm", "Pa", "pc", "ph", "photon",
            "pix", "pixel", "R", "rad", "Ry", "s", "S", "solLum", "solMass",
            "solRad", "sr", "T", "u", "V", "voxel", "W", "Wb", "yr",
        ]  # fmt: skip
        binary_bases = ["bit", "byte", "B"]
        simple_units = ["Angstrom", "angstrom", "AU", "au", "Ba", "dB", "mas", "Sun"]
        si_prefixes = [
            "y", "z", "a", "f", "p", "n", "u", "m", "c", "d",
            "", "da", "h", "k", "M", "G", "T", "P", "E", "Z", "Y"
        ]  # fmt: skip
        binary_prefixes = ["Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi"]
        deprecated_units = {"angstrom", "Angstrom", "Ba", "barn", "erg", "G", "ta"}

        def do_defines(bases, prefixes, skips=[]):
            for base in bases:
                for prefix in prefixes:
                    if (key := prefix + base) not in skips:
                        names[key] = getattr(u if hasattr(u, key) else uvo, key)
                        if base in deprecated_units:
                            deprecated_names.add(key)

        do_defines(bases, si_prefixes, ["pct", "pcount", "yd", "as"])
        do_defines(binary_bases, si_prefixes + binary_prefixes, ["dB", "dbyte"])
        do_defines(simple_units, [""])

        return names, frozenset(deprecated_names)

    @classproperty(lazy=True)
    def _units(cls) -> dict[str, UnitBase]:
        return cls._all_units[0]

    @classproperty(lazy=True)
    def _deprecated_units(cls) -> frozenset[str]:
        return cls._all_units[1]

    @classmethod
    def parse(cls, s: str, debug: bool = False) -> UnitBase:
        if s in ("unknown", "UNKNOWN"):
            return None
        if s == "":
            return dimensionless_unscaled
        # Check for excess solidi, but exclude fractional exponents (allowed)
        if s.count("/") > 1 and s.count("/") - len(re.findall(r"\(\d+/\d+\)", s)) > 1:
            raise UnitsError(
                f"'{s}' contains multiple slashes, which is "
                "disallowed by the VOUnit standard."
            )
        result = cls._do_parse(s, debug)
        if hasattr(result, "function_unit"):
            raise ValueError("Function units are not yet supported in VOUnit.")
        return result

    @classmethod
    def _get_unit(cls, t: LexToken) -> UnitBase:
        try:
            return super()._get_unit(t)
        except ValueError:
            if cls._explicit_custom_unit_regex.match(t.value):
                return cls._def_custom_unit(t.value)

            if cls._custom_unit_regex.match(t.value):
                warnings.warn(
                    cls._invalid_unit_error_message(t.value), UnitParserWarning
                )
                return cls._def_custom_unit(t.value)

            raise

    @classmethod
    def _decompose_to_known_units(
        cls,
        unit: CompositeUnit | NamedUnit,
        deprecations: DeprecatedUnitAction = DeprecatedUnitAction.WARN,
    ) -> UnitBase:
        # The da- and d- prefixes are discouraged.  This has the
        # effect of adding a scale to value in the result.
        if isinstance(unit, PrefixUnit) and unit._represents.scale in (0.1, 10.0):
            return cls._decompose_to_known_units(unit._represents)
        if (
            isinstance(unit, NamedUnit)
            and unit._get_format_name(cls.name) in cls._custom_units
        ):
            return unit
        return super()._decompose_to_known_units(unit, deprecations)

    @classmethod
    def _def_custom_unit(cls, unit: str) -> UnitBase:
        def def_base(name):
            if name in cls._custom_units:
                return cls._custom_units[name]

            if name.startswith("'"):
                return def_unit(
                    [name[1:-1], name],
                    format={"vounit": name},
                    namespace=cls._custom_units,
                )
            else:
                return def_unit(name, namespace=cls._custom_units)

        if unit in cls._custom_units:
            return cls._custom_units[unit]

        for short, _, factor in si_prefixes:
            for prefix in short:
                if unit.startswith(prefix):
                    base_name = unit[len(prefix) :]
                    base_unit = def_base(base_name)
                    return PrefixUnit(
                        [prefix + x for x in base_unit.names],
                        CompositeUnit(factor, [base_unit], [1], _error_check=False),
                        format={"vounit": prefix + base_unit.names[-1]},
                        namespace=cls._custom_units,
                    )

        return def_base(unit)

    @classmethod
    def _format_superscript(cls, number: str) -> str:
        return f"**({number})" if "/" in number or "." in number else f"**{number}"

    @classmethod
    def format_exponential_notation(
        cls, val: UnitScale | np.number, format_spec: str = ".8g"
    ) -> str:
        return format(val, format_spec)

    @classmethod
    def _format_inline_fraction(
        cls, scale: str, numerator: str, denominator: str
    ) -> str:
        if cls._space in denominator:
            denominator = f"({denominator})"
        if scale and numerator == "1":
            return f"{scale}/{denominator}"
        return f"{scale}{numerator}/{denominator}"

    @classmethod
    def to_string(
        cls,
        unit: UnitBase,
        fraction: bool | Literal["inline", "multiline"] = False,
        deprecations: DeprecatedUnitAction = DeprecatedUnitAction.WARN,
    ) -> str:
        # Remove units that aren't known to the format
        unit = cls._decompose_to_known_units(unit, deprecations)

        if unit.physical_type == "dimensionless" and unit.scale != 1:
            raise UnitScaleError(
                "The VOUnit format is not able to "
                "represent scale for dimensionless units. "
                f"Multiply your data by {unit.scale:e}."
            )

        return super().to_string(unit, fraction=fraction)

    @classmethod
    def _fix_deprecated(cls, x: str) -> list[str]:
        return (
            [f"{x} (deprecated)", cls.to_string(cls._units[x]._represents)]
            if x in cls._deprecated_units
            else [x]
        )



import re
import sys

import pytest

from astropy.io.votable.xmlutil import validate_schema
from astropy.utils.data import get_pkg_data_filename

# Each test file uses a feature that is new with its version.
V_1_2_TABLE = get_pkg_data_filename("data/empty_table.xml")
V_1_3_TABLE = get_pkg_data_filename("data/binary2_masked_strings.xml")
V_1_4_TABLE = get_pkg_data_filename("data/timesys.xml")
V_1_5_TABLE = get_pkg_data_filename("data/coosys.xml")

# Schema versions to test against
V_1_2 = "1.2"
V_1_3 = "1.3"
V_1_4 = "1.4"
V_1_5 = "1.5"
V_1_6 = "1.6"  # Next (unimplemented) schema

# Starting with version 1.4, schemas were upward compatible.
test_cases = [
    (V_1_2_TABLE, V_1_2, 0, None),
    (
        V_1_2_TABLE,
        V_1_3,
        3,
        b"No matching global declaration available for the validation root",
    ),
    (V_1_3_TABLE, V_1_3, 0, None),
    (V_1_3_TABLE, V_1_4, 0, None),
    (V_1_3_TABLE, V_1_5, 0, None),
    (
        V_1_3_TABLE,
        V_1_6,
        3,
        b"No matching global declaration available for the validation root",
    ),
    (V_1_4_TABLE, V_1_3, 3, b"TIMESYS.*element is not expected"),
    (V_1_4_TABLE, V_1_4, 0, None),
    (V_1_4_TABLE, V_1_5, 0, None),
    (
        V_1_4_TABLE,
        V_1_6,
        3,
        b"No matching global declaration available for the validation root",
    ),
    (V_1_5_TABLE, V_1_3, 3, b"attribute 'refposition' is not allowed"),
    (V_1_5_TABLE, V_1_4, 3, b"attribute 'refposition' is not allowed"),
    (V_1_5_TABLE, V_1_5, 0, None),
    (
        V_1_5_TABLE,
        V_1_6,
        3,
        b"No matching global declaration available for the validation root",
    ),
]


@pytest.mark.skipif(sys.platform.startswith("win"), reason="Cannot test on Windows")
@pytest.mark.parametrize(
    "votable_file,schema_version,expected_return_code,expected_msg_re", test_cases
)
def test_schema_versions(
    votable_file, schema_version, expected_return_code, expected_msg_re
):
    """Test that xmllint gives expected results for the given file and schema version."""
    try:
        rc, stdout, stderr = validate_schema(votable_file, schema_version)
    except OSError:
        # If xmllint is not installed, we want to skip the test
        pytest.skip("xmllint is not available so will not do schema validation")

    assert rc == expected_return_code
    if rc == 3:
        # Schema validation error.  Check the error message content.
        assert re.search(expected_msg_re, stderr)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__ = ["TotalComponent"]

from abc import abstractmethod

from numpy.typing import ArrayLike

from astropy.cosmology._src.typing import FArray
from astropy.units import Quantity


class TotalComponent:
    """The cosmology has attributes and methods for the total density.

    This trait has the abstract ``Otot`` method that returns the total density
    parameter at redshift ``z``. It should be the sum of all other components.
    """

    @property
    @abstractmethod
    def Otot0(self) -> float:
        """Omega total; the total density/critical density at z=0."""
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def Otot(self, z: Quantity | ArrayLike, /) -> FArray:
        """The total density parameter at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshifts.

        Returns
        -------
        Otot : array
            The total density relative to the critical density at each
            redshift.
        """
        raise NotImplementedError  # pragma: no cover



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np

from astropy.modeling import Parameter as ModelParameter
from astropy.table import Column

FULLQUALNAME_SUBSTITUTIONS = {
    "astropy.cosmology._src.flrw.base.FLRW": "astropy.cosmology.FLRW",
    "astropy.cosmology._src.flrw.lambdacdm.LambdaCDM": "astropy.cosmology.LambdaCDM",
    "astropy.cosmology._src.flrw.lambdacdm.FlatLambdaCDM": (
        "astropy.cosmology.FlatLambdaCDM"
    ),
    "astropy.cosmology._src.flrw.w0wacdm.w0waCDM": "astropy.cosmology.w0waCDM",
    "astropy.cosmology._src.flrw.w0wacdm.Flatw0waCDM": "astropy.cosmology.Flatw0waCDM",
    "astropy.cosmology._src.flrw.w0wzcdm.w0wzCDM": "astropy.cosmology.w0wzCDM",
    "astropy.cosmology._src.flrw.w0cdm.wCDM": "astropy.cosmology.wCDM",
    "astropy.cosmology._src.flrw.w0cdm.FlatwCDM": "astropy.cosmology.FlatwCDM",
    "astropy.cosmology._src.flrw.wpwazpcdm.wpwaCDM": "astropy.cosmology.wpwaCDM",
    # ===== deprecated paths =====
    "astropy.cosmology.flrw.base.FLRW": "astropy.cosmology.FLRW",
    "astropy.cosmology.flrw.lambdacdm.LambdaCDM": "astropy.cosmology.LambdaCDM",
    "astropy.cosmology.flrw.lambdacdm.FlatLambdaCDM": (
        "astropy.cosmology.flrw.FlatLambdaCDM"
    ),
    "astropy.cosmology.flrw.w0wacdm.w0waCDM": "astropy.cosmology.w0waCDM",
    "astropy.cosmology.flrw.w0wacdm.Flatw0waCDM": "astropy.cosmology.Flatw0waCDM",
    "astropy.cosmology.flrw.w0wzcdm.w0wzCDM": "astropy.cosmology.w0wzCDM",
    "astropy.cosmology.flrw.w0cdm.wCDM": "astropy.cosmology.wCDM",
    "astropy.cosmology.flrw.w0cdm.FlatwCDM": "astropy.cosmology.FlatwCDM",
    "astropy.cosmology.flrw.wpwazpcdm.wpwaCDM": "astropy.cosmology.wpwaCDM",
}
"""Substitutions mapping the actual qualified name to its preferred value."""


def convert_parameter_to_column(parameter, value, meta=None):
    """Convert a |Cosmology| Parameter to a Table |Column|.

    Parameters
    ----------
    parameter : `astropy.cosmology._src.parameter.Parameter`
    value : Any
    meta : dict or None, optional
        Information from the Cosmology's metadata.

    Returns
    -------
    `astropy.table.Column`
    """
    shape = (1,) + np.shape(value)  # minimum of 1d

    return Column(
        data=np.reshape(value, shape),
        name=parameter.name,
        dtype=None,  # inferred from the data
        description=parameter.__doc__,
        format=None,
        meta=meta,
    )


def convert_parameter_to_model_parameter(parameter, value, meta=None):
    """Convert a Cosmology Parameter to a Model Parameter.

    Parameters
    ----------
    parameter : `astropy.cosmology._src.parameter.Parameter`
    value : Any
    meta : dict or None, optional
        Information from the Cosmology's metadata.
        This function will use any of: 'getter', 'setter', 'fixed', 'tied',
        'min', 'max', 'bounds', 'prior', 'posterior'.

    Returns
    -------
    `astropy.modeling.Parameter`
    """
    # Get from meta information relevant to Model
    attrs = (
        "getter",
        "setter",
        "fixed",
        "tied",
        "min",
        "max",
        "bounds",
        "prior",
        "posterior",
    )
    extra = {k: v for k, v in (meta or {}).items() if k in attrs}

    return ModelParameter(
        description=parameter.__doc__,
        default=value,
        unit=getattr(value, "unit", None),
        **extra,
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import annotations

import os

from numpy import get_include as get_numpy_include
from setuptools import Extension

ROOT = os.path.relpath(os.path.dirname(__file__))
SRCFILES = ["wirth_select.c", "compute_bounds.c", "fast_sigma_clip.c"]
SRCFILES = [os.path.join(ROOT, "src", srcfile) for srcfile in SRCFILES]


def get_extensions() -> list[Extension, Extension]:
    _sigma_clip_ext = Extension(
        name="astropy.stats._fast_sigma_clip",
        sources=SRCFILES,
        include_dirs=[get_numpy_include()],
        language="c",
    )
    _stats_ext = Extension(
        name="astropy.stats._stats",
        sources=[os.path.join(ROOT, "_stats.pyx")],
        include_dirs=[get_numpy_include()],
    )

    return [_sigma_clip_ext, _stats_ext]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from copy import deepcopy

import numpy as np


def test_wcsprm_tab_basic(tab_wcs_2di):
    assert len(tab_wcs_2di.wcs.tab) == 1
    t = tab_wcs_2di.wcs.tab[0]
    assert tab_wcs_2di.wcs.tab[0] is not t


def test_tabprm_coord(tab_wcs_2di_f):
    t = tab_wcs_2di_f.wcs.tab[0]

    c0 = t.coord
    c1 = np.ones_like(c0)
    t.coord = c1
    assert np.allclose(tab_wcs_2di_f.wcs.tab[0].coord, c1)


def test_tabprm_crval_and_deepcopy(tab_wcs_2di_f):
    w = deepcopy(tab_wcs_2di_f)

    t = tab_wcs_2di_f.wcs.tab[0]

    pix = np.array([[2, 3]], dtype=np.float32)

    rd1 = tab_wcs_2di_f.wcs_pix2world(pix, 1)

    c = t.crval.copy()
    d = 0.5 * np.ones_like(c)
    t.crval += d
    assert np.allclose(tab_wcs_2di_f.wcs.tab[0].crval, c + d)

    rd2 = tab_wcs_2di_f.wcs_pix2world(pix - d, 1)
    assert np.allclose(rd1, rd2)

    rd3 = w.wcs_pix2world(pix, 1)
    assert np.allclose(rd1, rd3)


def test_tabprm_delta(tab_wcs_2di):
    t = tab_wcs_2di.wcs.tab[0]
    assert np.allclose([0.0, 0.0], t.delta)


def test_tabprm_K(tab_wcs_2di):
    t = tab_wcs_2di.wcs.tab[0]
    assert np.all(t.K == [4, 2])


def test_tabprm_M(tab_wcs_2di):
    t = tab_wcs_2di.wcs.tab[0]
    assert t.M == 2


def test_tabprm_nc(tab_wcs_2di):
    t = tab_wcs_2di.wcs.tab[0]
    assert t.nc == 8


def test_tabprm_extrema(tab_wcs_2di):
    t = tab_wcs_2di.wcs.tab[0]
    extrema = np.array(
        [
            [[-0.0026, -0.5], [1.001, -0.5]],
            [[-0.0026, 0.5], [1.001, 0.5]],
        ]
    )
    assert np.allclose(t.extrema, extrema)


def test_tabprm_map(tab_wcs_2di_f):
    t = tab_wcs_2di_f.wcs.tab[0]
    assert np.allclose(t.map, [0, 1])

    t.map[1] = 5
    assert np.all(tab_wcs_2di_f.wcs.tab[0].map == [0, 5])

    t.map = [1, 4]
    assert np.all(tab_wcs_2di_f.wcs.tab[0].map == [1, 4])


def test_tabprm_sense(tab_wcs_2di):
    t = tab_wcs_2di.wcs.tab[0]
    assert np.all(t.sense == [1, 1])


def test_tabprm_p0(tab_wcs_2di):
    t = tab_wcs_2di.wcs.tab[0]
    assert np.all(t.p0 == [0, 0])


def test_tabprm_print(tab_wcs_2di_f, capfd):
    tab_wcs_2di_f.wcs.tab[0].print_contents()
    captured = capfd.readouterr()
    s = str(tab_wcs_2di_f.wcs.tab[0])
    out = str(captured.out)
    lout = out.split("\n")
    assert out == s
    assert lout[0] == "       flag: 137"
    assert lout[1] == "          M: 2"


def test_wcstab_copy(tab_wcs_2di_f):
    t = tab_wcs_2di_f.wcs.tab[0]

    c0 = t.coord
    c1 = np.ones_like(c0)
    t.coord = c1
    assert np.allclose(tab_wcs_2di_f.wcs.tab[0].coord, c1)


def test_tabprm_crval(tab_wcs_2di_f):
    w = deepcopy(tab_wcs_2di_f)

    t = tab_wcs_2di_f.wcs.tab[0]

    pix = np.array([[2, 3]], dtype=np.float32)

    rd1 = tab_wcs_2di_f.wcs_pix2world(pix, 1)

    c = t.crval.copy()
    d = 0.5 * np.ones_like(c)
    t.crval += d
    assert np.allclose(tab_wcs_2di_f.wcs.tab[0].crval, c + d)

    rd2 = tab_wcs_2di_f.wcs_pix2world(pix - d, 1)
    assert np.allclose(rd1, rd2)

    rd3 = w.wcs_pix2world(pix, 1)
    assert np.allclose(rd1, rd3)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""|Cosmology| <-> Mapping I/O, using |Cosmology.to_format| and |Cosmology.from_format|.

This module provides functions to transform a |Cosmology| instance to a mapping
(`dict`-like) object and vice versa, from a mapping object back to a |Cosmology|
instance. The functions are registered with ``convert_registry`` under the format name
"mapping". The mapping object is a `dict`-like object, with the cosmology's parameters
and metadata as items. `dict` is a fundamental data structure in Python, and this
representation of a |Cosmology| is useful for translating between many serialization and
storage formats, or even passing arguments to functions.

We start with the simple case of outputting a |Cosmology| as a mapping.

    >>> from astropy.cosmology import Cosmology, Planck18
    >>> cm = Planck18.to_format('mapping')
    >>> cm
    {'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>,
     'name': 'Planck18', 'H0': <Quantity 67.66 km / (Mpc s)>, 'Om0': 0.30966,
     'Tcmb0': <Quantity 2.7255 K>, 'Neff': 3.046,
     'm_nu': <Quantity [0. , 0. , 0.06] eV>, 'Ob0': 0.04897,
     'meta': ...

``cm`` is a `dict`, with the cosmology's parameters and metadata as items.

How might we use this `dict`? One use is to unpack the `dict` into a function:

    >>> def function(H0, Tcmb0, **kwargs): ...
    >>> function(**cm)

Another use is to merge the `dict` with another `dict`:

    >>> cm2 = {'H0': 70, 'Tcmb0': 2.7}
    >>> cm | cm2
    {'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>, ..., 'H0': 70, ...}

Most saliently, the `dict` can also be used to construct a new cosmological instance
identical to the |Planck18| cosmology from which it was generated.

    >>> cosmo = Cosmology.from_format(cm, format="mapping")
    >>> cosmo
    FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

How did |Cosmology.from_format| know to return an instance of the |FlatLambdaCDM| class?
The mapping object has a field ``cosmology`` which can be either the string name of the
cosmology class (e.g. "FlatLambdaCDM") or the class itself.

This field can be omitted under two conditions.

1. If the cosmology class is passed as the ``cosmology`` keyword argument to
   |Cosmology.from_format|,
2. If a specific cosmology class, e.g. |FlatLambdaCDM|, is used to parse the data.

To the first point, we can pass the cosmology class as the ``cosmology`` keyword
argument to |Cosmology.from_format|.

    >>> del cm["cosmology"]  # remove cosmology class

    >>> Cosmology.from_format(cm, cosmology="FlatLambdaCDM")
    FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

To the second point, we can use specific cosmology class to parse the data.

    >>> from astropy.cosmology import FlatLambdaCDM
    >>> FlatLambdaCDM.from_format(cm)
    FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

Also, the class' default parameter values are used to fill in any information missing in
the data. For example, if ``Tcmb0`` is missing, the default value of 0.0 K is used.

    >>> del cm["Tcmb0"]  # show FlatLambdaCDM provides default
    >>> FlatLambdaCDM.from_format(cm)
    FlatLambdaCDM(name='Planck18', H0=..., Tcmb0=<Quantity 0. K>, ...)

If instead of *missing* information, there is *extra* information, there are a few
options. The first is to use the ``move_to_meta`` keyword argument to move fields that
are not in the Cosmology constructor to the Cosmology's metadata.

    >>> cm2 = cm | {"extra": 42, "cosmology": "FlatLambdaCDM"}
    >>> cosmo = Cosmology.from_format(cm2, move_to_meta=True)
    >>> cosmo.meta
    {'extra': 42, ...}

Alternatively, the ``rename`` keyword argument can be used to rename keys in the mapping
to fields of the |Cosmology|. This is crucial when the mapping has keys that are not
valid arguments to the |Cosmology| constructor.

    >>> cm3 = dict(cm)  # copy
    >>> cm3["cosmo_cls"] = "FlatLambdaCDM"
    >>> cm3["cosmo_name"] = cm3.pop("name")

    >>> rename = {'cosmo_cls': 'cosmology', 'cosmo_name': 'name'}
    >>> Cosmology.from_format(cm3, rename=rename)
    FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 0. K>, Neff=3.046, m_nu=None, Ob0=0.04897)

Let's take a closer look at |Cosmology.to_format|, because there a lot of options, to
tailor the output to specific needs.

The dictionary type may be changed with the ``cls`` keyword argument:

    >>> from collections import OrderedDict
    >>> Planck18.to_format('mapping', cls=OrderedDict)
    OrderedDict({'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>, 'name': 'Planck18', 'H0': <Quantity 67.66 km / (Mpc s)>, 'Om0': 0.30966, 'Tcmb0': <Quantity 2.7255 K>, 'Neff': 3.046, 'm_nu': <Quantity [0.  , 0.  , 0.06] eV>, 'Ob0': 0.04897, 'meta': {...}})

Sometimes it is more useful to have the name of the cosmology class, not the type
itself. The keyword argument ``cosmology_as_str`` may be used:

    >>> Planck18.to_format('mapping', cosmology_as_str=True)
    {'cosmology': 'FlatLambdaCDM', ...

The metadata is normally included as a nested mapping. To move the metadata into the
main mapping, use the keyword argument ``move_from_meta``. This kwarg inverts
``move_to_meta`` in ``Cosmology.to_format("mapping", move_to_meta=...)`` where extra
items are moved to the metadata (if the cosmology constructor does not have a variable
keyword-only argument -- ``**kwargs``).

    >>> from astropy.cosmology import Planck18
    >>> Planck18.to_format('mapping', move_from_meta=True)
    {'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>,
        'name': 'Planck18', 'Oc0': 0.2607, 'n': 0.9665, 'sigma8': 0.8102, ...

Lastly, the keys in the mapping may be renamed with the ``rename`` keyword.

    >>> rename = {'cosmology': 'cosmo_cls', 'name': 'cosmo_name'}
    >>> Planck18.to_format('mapping', rename=rename)
    {'cosmo_cls': <class 'astropy.cosmology...FlatLambdaCDM'>,
     'cosmo_name': 'Planck18', ...
"""

__all__: list[str] = []  # nothing is publicly scoped

import copy
import inspect
from collections.abc import Mapping, MutableMapping
from typing import Any, TypeVar

from astropy.cosmology._src.core import _COSMOLOGY_CLASSES, Cosmology
from astropy.cosmology._src.io.connect import convert_registry
from astropy.cosmology._src.typing import _CosmoT

_MapT = TypeVar("_MapT", bound=MutableMapping[str, Any])


def _rename_map(
    map: Mapping[str, Any], /, renames: Mapping[str, str]
) -> dict[str, Any]:
    """Apply rename to map."""
    if common_names := set(renames.values()).intersection(map):
        raise ValueError(
            "'renames' values must be disjoint from 'map' keys, "
            f"the common keys are: {common_names}"
        )
    return {renames.get(k, k): v for k, v in map.items()}  # dict separate from input


def _get_cosmology_class(
    cosmology: type[_CosmoT] | str | None, params: dict[str, Any], /
) -> type[_CosmoT]:
    # get cosmology
    # 1st from argument. Allows for override of the cosmology, if on file.
    # 2nd from params. This MUST have the cosmology if 'kwargs' did not.
    if cosmology is None:
        cosmology = params.pop("cosmology")
    else:
        params.pop("cosmology", None)  # pop, but don't use
    # if string, parse to class
    return _COSMOLOGY_CLASSES[cosmology] if isinstance(cosmology, str) else cosmology


def from_mapping(
    mapping: Mapping[str, Any],
    /,
    *,
    move_to_meta: bool = False,
    cosmology: str | type[_CosmoT] | None = None,
    rename: Mapping[str, str] | None = None,
) -> _CosmoT:
    """Load `~astropy.cosmology.Cosmology` from mapping object.

    Parameters
    ----------
    mapping : Mapping
        Arguments into the class -- like "name" or "meta". If 'cosmology' is None, must
        have field "cosmology" which can be either the string name of the cosmology
        class (e.g. "FlatLambdaCDM") or the class itself.

    move_to_meta : bool (optional, keyword-only)
        Whether to move keyword arguments that are not in the Cosmology class' signature
        to the Cosmology's metadata. This will only be applied if the Cosmology does NOT
        have a keyword-only argument (e.g. ``**kwargs``). Arguments moved to the
        metadata will be merged with existing metadata, preferring specified metadata in
        the case of a merge conflict (e.g. for ``Cosmology(meta={'key':10}, key=42)``,
        the ``Cosmology.meta`` will be ``{'key': 10}``).

    cosmology : str, |Cosmology| class, or None (optional, keyword-only)
        The cosmology class (or string name thereof) to use when constructing the
        cosmology instance. The class also provides default parameter values, filling in
        any non-mandatory arguments missing in 'map'.

    rename : Mapping[str, str] or None (optional, keyword-only)
        A mapping of keys in ``map`` to fields of the `~astropy.cosmology.Cosmology`.

    Returns
    -------
    `~astropy.cosmology.Cosmology` subclass instance

    Examples
    --------
    To see loading a `~astropy.cosmology.Cosmology` from a dictionary with
    ``from_mapping``, we will first make a mapping using
    :meth:`~astropy.cosmology.Cosmology.to_format`.

        >>> from astropy.cosmology import Cosmology, Planck18
        >>> cm = Planck18.to_format('mapping')
        >>> cm
        {'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>,
         'name': 'Planck18', 'H0': <Quantity 67.66 km / (Mpc s)>, 'Om0': 0.30966,
         'Tcmb0': <Quantity 2.7255 K>, 'Neff': 3.046,
         'm_nu': <Quantity [0. , 0. , 0.06] eV>, 'Ob0': 0.04897,
         'meta': ...

    Now this dict can be used to load a new cosmological instance identical to the
    |Planck18| cosmology from which it was generated.

        >>> cosmo = Cosmology.from_format(cm, format="mapping")
        >>> cosmo
        FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

    The ``cosmology`` field can be omitted if the cosmology class (or its string name)
    is passed as the ``cosmology`` keyword argument to |Cosmology.from_format|.

        >>> del cm["cosmology"]  # remove cosmology class
        >>> Cosmology.from_format(cm, cosmology="FlatLambdaCDM")
        FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

    Alternatively, specific cosmology classes can be used to parse the data.

        >>> from astropy.cosmology import FlatLambdaCDM
        >>> FlatLambdaCDM.from_format(cm)
        FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 2.7255 K>, Neff=3.046, m_nu=<Quantity [0.  , 0.  , 0.06] eV>, Ob0=0.04897)

    When using a specific cosmology class, the class' default parameter values are used
    to fill in any missing information.

        >>> del cm["Tcmb0"]  # show FlatLambdaCDM provides default
        >>> FlatLambdaCDM.from_format(cm)
        FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 0. K>, Neff=3.046, m_nu=None, Ob0=0.04897)

    The ``move_to_meta`` keyword argument can be used to move fields that are not in the
    Cosmology constructor to the Cosmology's metadata. This is useful when the
    dictionary contains extra information that is not part of the Cosmology.

        >>> cm2 = cm | {"extra": 42, "cosmology": "FlatLambdaCDM"}
        >>> cosmo = Cosmology.from_format(cm2, move_to_meta=True)
        >>> cosmo.meta
        {'extra': 42, ...}

    The ``rename`` keyword argument can be used to rename keys in the mapping to fields
    of the |Cosmology|. This is crucial when the mapping has keys that are not valid
    arguments to the |Cosmology| constructor.

        >>> cm3 = dict(cm)  # copy
        >>> cm3["cosmo_cls"] = "FlatLambdaCDM"
        >>> cm3["cosmo_name"] = cm3.pop("name")

        >>> rename = {'cosmo_cls': 'cosmology', 'cosmo_name': 'name'}
        >>> Cosmology.from_format(cm3, rename=rename)
        FlatLambdaCDM(name='Planck18', H0=<Quantity 67.66 km / (Mpc s)>, Om0=0.30966, Tcmb0=<Quantity 0. K>, Neff=3.046, m_nu=None, Ob0=0.04897)
    """
    # Rename keys, if given a ``renames`` dict.
    # Also, make a copy of the mapping, so we can pop from it.
    params = _rename_map(dict(mapping), renames=rename or {})

    # Get cosmology class
    cosmology = _get_cosmology_class(cosmology, params)

    # select arguments from mapping that are in the cosmo's signature.
    sig = inspect.signature(cosmology)
    ba = sig.bind_partial()  # blank set of args
    ba.apply_defaults()  # fill in the defaults
    for k in sig.parameters.keys():
        if k in params:  # transfer argument, if in params
            ba.arguments[k] = params.pop(k)

    # deal with remaining params. If there is a **kwargs use that, else
    # allow to transfer to metadata. Raise TypeError if can't.
    lastp = next(reversed(sig.parameters.values()))
    if lastp.kind == 4:  # variable keyword-only
        ba.arguments[lastp.name] = params
    elif move_to_meta:  # prefers current meta, which was explicitly set
        meta = ba.arguments["meta"] or {}  # (None -> dict)
        ba.arguments["meta"] = {**params, **meta}
    elif params:
        raise TypeError(f"there are unused parameters {params}.")
    # else: pass  # no kwargs, no move-to-meta, and all the params are used

    return cosmology(*ba.args, **ba.kwargs)


def to_mapping(
    cosmology: Cosmology,
    *args: object,
    cls: type[_MapT] = dict,
    cosmology_as_str: bool = False,
    move_from_meta: bool = False,
    rename: Mapping[str, str] | None = None,
) -> _MapT:
    """Return the cosmology class, parameters, and metadata as a `dict`.

    Parameters
    ----------
    cosmology : :class:`~astropy.cosmology.Cosmology`
        The cosmology instance to convert to a mapping.
    *args : object
        Not used. Needed for compatibility with
        `~astropy.io.registry.UnifiedReadWriteMethod`
    cls : type (optional, keyword-only)
        `dict` or `collections.Mapping` subclass.
        The mapping type to return. Default is `dict`.
    cosmology_as_str : bool (optional, keyword-only)
        Whether the cosmology value is the class (if `False`, default) or
        the semi-qualified name (if `True`).
    move_from_meta : bool (optional, keyword-only)
        Whether to add the Cosmology's metadata as an item to the mapping (if
        `False`, default) or to merge with the rest of the mapping, preferring
        the original values (if `True`)
    rename : Mapping[str, str] or None (optional, keyword-only)
        A mapping of field names of the :class:`~astropy.cosmology.Cosmology` to keys in
        the map.

    Returns
    -------
    MutableMapping[str, Any]
        A mapping of type ``cls``, by default a `dict`.
        Has key-values for the cosmology parameters and also:
        - 'cosmology' : the class
        - 'meta' : the contents of the cosmology's metadata attribute.
                   If ``move_from_meta`` is `True`, this key is missing and the
                   contained metadata are added to the main `dict`.

    Examples
    --------
    A Cosmology as a mapping will have the cosmology's name and
    parameters as items, and the metadata as a nested dictionary.

        >>> from astropy.cosmology import Planck18
        >>> Planck18.to_format('mapping')
        {'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>,
         'name': 'Planck18', 'H0': <Quantity 67.66 km / (Mpc s)>, 'Om0': 0.30966,
         'Tcmb0': <Quantity 2.7255 K>, 'Neff': 3.046,
         'm_nu': <Quantity [0.  , 0.  , 0.06] eV>, 'Ob0': 0.04897,
         'meta': ...

    The dictionary type may be changed with the ``cls`` keyword argument:

        >>> from collections import OrderedDict
        >>> Planck18.to_format('mapping', cls=OrderedDict)
        OrderedDict({'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>, 'name': 'Planck18', 'H0': <Quantity 67.66 km / (Mpc s)>, 'Om0': 0.30966, 'Tcmb0': <Quantity 2.7255 K>, 'Neff': 3.046, 'm_nu': <Quantity [0.  , 0.  , 0.06] eV>, 'Ob0': 0.04897, 'meta': {...}})

    Sometimes it is more useful to have the name of the cosmology class, not
    the type itself. The keyword argument ``cosmology_as_str`` may be used:

        >>> Planck18.to_format('mapping', cosmology_as_str=True)
        {'cosmology': 'FlatLambdaCDM', ...

    The metadata is normally included as a nested mapping. To move the metadata
    into the main mapping, use the keyword argument ``move_from_meta``. This
    kwarg inverts ``move_to_meta`` in
    ``Cosmology.to_format("mapping", move_to_meta=...)`` where extra items
    are moved to the metadata (if the cosmology constructor does not have a
    variable keyword-only argument -- ``**kwargs``).

        >>> from astropy.cosmology import Planck18
        >>> Planck18.to_format('mapping', move_from_meta=True)
        {'cosmology': <class 'astropy.cosmology...FlatLambdaCDM'>,
         'name': 'Planck18', 'Oc0': 0.2607, 'n': 0.9665, 'sigma8': 0.8102, ...

    Lastly, the keys in the mapping may be renamed with the ``rename`` keyword.

        >>> rename = {'cosmology': 'cosmo_cls', 'name': 'cosmo_name'}
        >>> Planck18.to_format('mapping', rename=rename)
        {'cosmo_cls': <class 'astropy.cosmology...FlatLambdaCDM'>,
         'cosmo_name': 'Planck18', ...
    """
    if not issubclass(cls, (dict, Mapping)):
        raise TypeError(f"'cls' must be a (sub)class of dict or Mapping, not {cls}")

    m = cls()
    # start with the cosmology class & name
    m["cosmology"] = (
        cosmology.__class__.__qualname__ if cosmology_as_str else cosmology.__class__
    )
    m["name"] = cosmology.name  # here only for dict ordering

    meta = copy.deepcopy(cosmology.meta)  # metadata (mutable)
    if move_from_meta:
        # Merge the mutable metadata. Since params are added later they will
        # be preferred in cases of overlapping keys. Likewise, need to pop
        # cosmology and name from meta.
        meta.pop("cosmology", None)
        meta.pop("name", None)
        m.update(meta)

    # Add all the immutable inputs
    m.update(cosmology.parameters)
    # Lastly, add the metadata, if haven't already (above)
    if not move_from_meta:
        m["meta"] = meta  # TODO? should meta be type(cls)
    # Rename keys
    return m if rename is None else _rename_map(m, rename)


def mapping_identify(
    origin: str, format: str | None, *args: object, **kwargs: object
) -> bool:
    """Identify if object uses the mapping format.

    Returns
    -------
    bool
    """
    itis = False
    if origin == "read":
        itis = isinstance(args[1], Mapping) and (format in (None, "mapping"))
    return itis


# ===================================================================
# Register

convert_registry.register_reader("mapping", Cosmology, from_mapping)
convert_registry.register_writer("mapping", Cosmology, to_mapping)
convert_registry.register_identifier("mapping", Cosmology, mapping_identify)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""URL unescaper functions."""

# STDLIB
from xml.sax import saxutils

__all__ = ["unescape_all"]

# This is DIY
_bytes_entities = {
    b"&amp;": b"&",
    b"&lt;": b"<",
    b"&gt;": b">",
    b"&amp;&amp;": b"&",
    b"&&": b"&",
    b"%2F": b"/",
}
_bytes_keys = [b"&amp;&amp;", b"&&", b"&amp;", b"&lt;", b"&gt;", b"%2F"]

# This is used by saxutils
_str_entities = {"&amp;&amp;": "&", "&&": "&", "%2F": "/"}
_str_keys = ["&amp;&amp;", "&&", "&amp;", "&lt;", "&gt;", "%2F"]


def unescape_all(url):
    """Recursively unescape a given URL.

    .. note:: '&amp;&amp;' becomes a single '&'.

    Parameters
    ----------
    url : str or bytes
        URL to unescape.

    Returns
    -------
    clean_url : str or bytes
        Unescaped URL.

    """
    if isinstance(url, bytes):
        func2use = _unescape_bytes
        keys2use = _bytes_keys
    else:
        func2use = _unescape_str
        keys2use = _str_keys
    clean_url = func2use(url)
    not_done = [clean_url.count(key) > 0 for key in keys2use]
    if True in not_done:
        return unescape_all(clean_url)
    else:
        return clean_url


def _unescape_str(url):
    return saxutils.unescape(url, _str_entities)


def _unescape_bytes(url):
    clean_url = url
    for key in _bytes_keys:
        clean_url = clean_url.replace(key, _bytes_entities[key])
    return clean_url



# Licensed under a 3-clause BSD style license - see LICENSE.rst

__all__ = [  # noqa: F822
    "CosmologyFromFormat",
    "CosmologyRead",
    "CosmologyToFormat",
    "CosmologyWrite",
]

import sys
import warnings
from typing import Any


def __getattr__(name: str) -> Any:
    if name not in __all__ + ["convert_registry", "readwrite_registry"]:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}.")

    from ._src.io import connect

    obj = getattr(connect, name)

    setattr(sys.modules[__name__], name, obj)

    warnings.warn(
        "The module `astropy.cosmology.connect` is deprecated since v7.1 and will be "
        "removed in a future version. Import from `astropy.cosmology.io` instead.",
        category=DeprecationWarning,
    )

    return obj



# Licensed under a 3-clause BSD style license - see PYFITS.rst

import gzip
import io

from astropy.io.fits.file import _File
from astropy.io.fits.header import Header, _pad_length
from astropy.io.fits.util import fileobj_name
from astropy.utils import lazyproperty

from .base import NonstandardExtHDU
from .hdulist import HDUList


class FitsHDU(NonstandardExtHDU):
    """
    A non-standard extension HDU for encapsulating entire FITS files within a
    single HDU of a container FITS file.  These HDUs have an extension (that is
    an XTENSION keyword) of FITS.

    The FITS file contained in the HDU's data can be accessed by the `hdulist`
    attribute which returns the contained FITS file as an `HDUList` object.
    """

    _extension = "FITS"

    @lazyproperty
    def hdulist(self):
        self._file.seek(self._data_offset)
        fileobj = io.BytesIO()
        # Read the data into a BytesIO--reading directly from the file
        # won't work (at least for gzipped files) due to problems deep
        # within the gzip module that make it difficult to read gzip files
        # embedded in another file
        fileobj.write(self._file.read(self.size))
        fileobj.seek(0)
        if self._header["COMPRESS"]:
            fileobj = gzip.GzipFile(fileobj=fileobj)
        return HDUList.fromfile(fileobj, mode="readonly")

    @classmethod
    def fromfile(cls, filename, compress=False):
        """
        Like `FitsHDU.fromhdulist()`, but creates a FitsHDU from a file on
        disk.

        Parameters
        ----------
        filename : str
            The path to the file to read into a FitsHDU
        compress : bool, optional
            Gzip compress the FITS file
        """
        with HDUList.fromfile(filename) as hdulist:
            return cls.fromhdulist(hdulist, compress=compress)

    @classmethod
    def fromhdulist(cls, hdulist, compress=False):
        """
        Creates a new FitsHDU from a given HDUList object.

        Parameters
        ----------
        hdulist : HDUList
            A valid Headerlet object.
        compress : bool, optional
            Gzip compress the FITS file
        """
        fileobj = bs = io.BytesIO()
        if compress:
            if hasattr(hdulist, "_file"):
                name = fileobj_name(hdulist._file)
            else:
                name = None
            fileobj = gzip.GzipFile(name, mode="wb", fileobj=bs)

        hdulist.writeto(fileobj)

        if compress:
            fileobj.close()

        # A proper HDUList should still be padded out to a multiple of 2880
        # technically speaking
        padding = (_pad_length(bs.tell()) * cls._padding_byte).encode("ascii")
        bs.write(padding)

        bs.seek(0)

        cards = [
            ("XTENSION", cls._extension, "FITS extension"),
            ("BITPIX", 8, "array data type"),
            ("NAXIS", 1, "number of array dimensions"),
            ("NAXIS1", len(bs.getvalue()), "Axis length"),
            ("PCOUNT", 0, "number of parameters"),
            ("GCOUNT", 1, "number of groups"),
        ]

        # Add the XINDn keywords proposed by Perry, though nothing is done with
        # these at the moment
        if len(hdulist) > 1:
            for idx, hdu in enumerate(hdulist[1:]):
                cards.append(
                    (
                        "XIND" + str(idx + 1),
                        hdu._header_offset,
                        f"byte offset of extension {idx + 1}",
                    )
                )

        cards.append(("COMPRESS", compress, "Uses gzip compression"))
        header = Header(cards)
        return cls._readfrom_internal(_File(bs), header=header)

    @classmethod
    def match_header(cls, header):
        card = header.cards[0]
        if card.keyword != "XTENSION":
            return False
        xtension = card.value
        if isinstance(xtension, str):
            xtension = xtension.rstrip()
        return xtension == cls._extension

    # TODO: Add header verification

    def _summary(self):
        # TODO: Perhaps make this more descriptive...
        return (self.name, self.ver, self.__class__.__name__, len(self._header))
